# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
r"""Greedy Fitting Module."""

from abc import ABCMeta, abstractmethod

import numpy as np
from scipy.optimize import nnls

from bfit.fit import _BaseFit, KLDivergenceFPI, ScipyFit
from bfit.measure import SquaredDifference
from bfit.model import AtomicGaussianDensity

__all__ = ["GreedyLeastSquares", "GreedyKLFPI"]


def remove_redundancies(coeffs, fparams, eps=1e-3):
    r"""
    Check if the exponents have similar values and group them together.

    If any two function parameters have similar values, then one of the
    function parameters is removed, and the corresponding coefficient,
    are added together.
    Note: as of now this only works if each basis function depends on only one
    parameters e.g. e^(-x), not e^(-x + y).

    Parameters
    ----------
    coeffs : np.ndarray(M,)
        Coefficients of the basis function set of size :math:`M`.
    fparams : np.ndarray(M,)
        Function parameters of the basis function set of size :math:`M`.
    eps : float
        Value that indicates the threshold for how close two parameters are.

    Returns
    -------
    (np.ndarray, np.ndarray) :
        New coefficients and new function parameters, where close values of the function parameters
        are removed and the coefficients corresponding to that parameter are added together.

    """
    new_coeffs = coeffs.copy()
    new_exps = fparams.copy()

    # Indexes where they're the same.
    indexes_same = []
    for i, alpha in enumerate(fparams):
        similar_indexes = []
        for j in range(i + 1, len(fparams)):
            if j not in similar_indexes:
                if np.abs(alpha - fparams[j]) < eps:
                    if i not in similar_indexes:
                        similar_indexes.append(i)
                    similar_indexes.append(j)
        if len(similar_indexes) != 0:
            indexes_same.append(similar_indexes)

    # Add the coefficients together and add/group them, if need be
    for group_similar_items in indexes_same:
        for i in range(1, len(group_similar_items)):
            new_coeffs[group_similar_items[0]] += coeffs[group_similar_items[i]]

    if len(indexes_same) != 0:
        indices = [y for x in indexes_same for y in x[1:]]
        new_exps = np.delete(new_exps, indices)
        new_coeffs = np.delete(new_coeffs, indices)
    return new_coeffs, new_exps


def get_next_choices(factor, coeffs, fparams, coeff_val=100.):
    r"""
    Get the next set of (n+1) fparams, used by the greedy-fitting algorithm.

    Assuming coeffs=:math:`\[c1, c2 ,... ,cn\]` and fparams =:math:`\[a1, a2, ..., an\]`,
    this method gets the next (n+1) fparams by using a constant factor
    and coefficient value :math:`c^\prime`. A list of (n+1) choices are
    returned. They are determined as follows:

    .. math::
        \begin{align*}
            \[a_1 / factor, a_2,  .., a_n\] &,\quad coeffs = [c1, c^\prime, c2, ..., cn], \\
            \[[a_1, a_2,...,(a_i + a_{(i+1)/2}, a_{i+1}, ..., an\] &,\quad \text{similar coeffs},\\
            \[[a_1, a_2, ..., factor * a_n\] &,\quad \text{similar coeffs},
        \end{align*}

    Parameters
    ----------
    factor : float
        Number used to give two choices by multiplying each end point.
    coeffs : np.ndarray
        Coefficients of the basis functions.
    fparams : np.ndarray
        Function parameters.
    coeff_val : float
        Number used to fill in the coefficient value for each guess.

    Returns
    -------
    List[np.ndarray]
        List of the next possible initial guesses for :math:`(n+1)` basis-functions,
        coefficients are listed first, then exponents.
    """
    size = fparams.shape[0]
    all_choices = []
    for index, exp in np.ndenumerate(fparams):
        if index[0] == 0:
            exps_arr = np.insert(fparams, index, exp / factor)
            coeffs_arr = np.insert(coeffs, index, coeff_val)
        elif index[0] <= size:
            exps_arr = np.insert(fparams, index, (fparams[index[0] - 1] +
                                                  fparams[index[0]]) / 2)
            coeffs_arr = np.insert(coeffs, index, coeff_val)
        all_choices.append(np.append(coeffs_arr, exps_arr))
        if index[0] == size - 1:
            exps_arr = np.append(fparams, np.array([exp * factor]))
            endpt = np.append(coeffs, np.array([coeff_val]))
            all_choices.append(np.append(endpt, exps_arr))
    return all_choices


def get_two_next_choices(factor, coeffs, fparams, coeff_val=100.):
    r"""
    Return a list of (n+2) set of initial guess for fparams for greedy.

    Assuming coeffs=:math:`\[c1, c2 ,... ,cn\]` and fparams=:math:`\[a1, a2, ..., an\]`.
    The next (n+2) choice is either a combination of an endpoint guess and a
    midpoint, or two mid point guess or two endpoint guess.
    This is two-times recursion of the function `get_next_choices`.

    .. math::
        \begin{align*}
            \[a_1 / factor, ..., (a_i + a_{(i+1))/2}, ..., a_n\],
            \[a_1, ..., (a_j + a_{(j+1))/2}, a_{j+1} ..., (a_i + a_{(i+1))/2}, a_{i+1}, .., a_n\],
            \[a_1 / factor, a_2, a_3, ..., a_{n-1}, factor * a_n\]
        \end{align*}

    Parameters
    ----------
    factor : float
        Number used to give two choices by multiplying each end point.
    coeffs : np.ndarray(M,)
        Coefficients of the basis function set of size :math:`M`.
    fparams : np.ndarray(M,)
        Function parameters of the basis function set of size :math:`M`.
    coeff_val : float
        Number used to fill in the coefficient value for each guess.

    Returns
    -------
    List[np.ndarray]
        List of the next possible initial guesses for :math:`(n+2)` basis-functions,
        coefficients are listed first, then exponents.

    """

    def _get_next_possible_coeffs_and_exps2(factor, coeffs, exps, coeff_val=100.):
        # This is the same function as get_next_choices except that
        #   it returns the coefficients and exponents separately.
        size = exps.shape[0]
        all_choices_of_exponents = []
        all_choices_of_coeffs = []
        for index, exp in np.ndenumerate(exps):
            if index[0] == 0:
                exponent_array = np.insert(exps, index, exp / factor)
                coefficient_array = np.insert(coeffs, index, coeff_val)
            elif index[0] <= size:
                exponent_array = np.insert(exps, index, (exps[index[0] - 1] + exps[index[0]]) / 2)
                coefficient_array = np.insert(coeffs, index, coeff_val)
            all_choices_of_exponents.append(exponent_array)
            all_choices_of_coeffs.append(coefficient_array)
            if index[0] == size - 1:
                exponent_array = np.append(exps, np.array([exp * factor]))
                all_choices_of_exponents.append(exponent_array)
                all_choices_of_coeffs.append(np.append(coeffs, np.array([coeff_val])))
        return all_choices_of_coeffs, all_choices_of_exponents

    size = len(fparams)
    choices_coeff = []
    choices_fparams = []
    for i, e in enumerate(fparams):
        if i == 0:
            fparam_arr = np.insert(fparams, 0, e / factor)
            coeff_arr = np.insert(coeffs, 0, coeff_val)
        elif i <= size:
            fparam_arr = np.insert(fparams, i, (fparams[i - 1] + fparams[i]) / 2)
            coeff_arr = np.insert(coeffs, i, coeff_val)

        coeff2, exp2 = _get_next_possible_coeffs_and_exps2(factor, coeff_arr, fparam_arr, coeff_val)
        choices_coeff.extend(coeff2[i:])
        choices_fparams.extend(exp2[i:])

        if i == size - 1:
            fparam_arr = np.append(fparams, np.array([e * factor]))
            endpt = np.append(coeffs, np.array([coeff_val]))
            coeff2, exp2 = _get_next_possible_coeffs_and_exps2(factor, endpt, fparam_arr, coeff_val)
            choices_coeff.extend(coeff2[-2:])
            choices_fparams.extend(exp2[-2:])

    # Append coefficients and exponents together.
    all_choices_params = []
    for i, c in enumerate(choices_coeff):
        all_choices_params.append(np.append(c, choices_fparams[i]))
    return all_choices_params


def pick_two_lose_one(factor, coeffs, exps, coeff_val=100.):
    r"""
    Get (n+1) initial guesses by choosing (n+2) guesses and losing one value each time.

    Most accurate out of the other methods in this file but has returns
    a large number of possible (n+1) initial guesses for greedy-algorithm.

    Using the choices from 'get_two_next_choices', this
    method removes each possible exponent, to get another (n+1) choice.

    Parameters
    ----------
    factor : float
        Number used to give two choices by multiplying each end point.
    coeffs : np.ndarray(M,)
        Coefficients of the basis function set of size :math:`M`.
    fparams : np.ndarray(M,)
        Function parameters of the basis function set of size :math:`M`.
    coeff_val : float
        Number used to fill in the coefficient value for each guess.

    Returns
    -------
    List[np.ndarray]
        List of the next possible initial guesses for `(n+1)` basis-functions,
        coefficients are listed first, then exponents.
    """
    all_choices = []
    two_choices = get_two_next_choices(factor, coeffs, exps, coeff_val)
    for p in two_choices:
        coeff, exp = p[:len(p) // 2], p[len(p) // 2:]
        for j in range(0, len(p) // 2):
            new_coeff = np.delete(coeff, j)
            new_exp = np.delete(exp, j)
            all_choices.append(np.append(new_coeff, new_exp))
    return all_choices


class GreedyStrategy(metaclass=ABCMeta):
    r"""Base greedy strategy class for fitting s-type, p-type Gaussians."""

    def __init__(self, fitting_obj, choice_function="pick-one"):
        r"""
        Construct the base greedy class.

        Parameters
        ----------
        choice_function : str
            Determines how the next set of basis-functions are chosen.
            Can be either:

            * `pick-one` : add a new basis-function by taking the average between every two
              s-type or p-type exponents.
            * `pick-two` : add two new basis-functions by recursion of "pick-one" over each
              guess from `pick-one`.
            * `pick-two-lose-one` : add new basis-function by iterating though each guess
              in `pick-two` and removing one basis-function, generating a new guess each time.
        fitting_obj : _BaseFit
            The fitting class that is a child of _BaseFit class.  This is used to
            optimize the objective function.

        """
        if not isinstance(fitting_obj, _BaseFit):
            raise TypeError(f"Fitting object {type(fitting_obj)} should be of type _BaseFit.")

        if choice_function == "pick-one":
            self.next_param_func = get_next_choices
            self._numb_func_increase = 1  # How many basis functions were increased
        elif choice_function == "pick-two":
            self.next_param_func = get_two_next_choices
            self._numb_func_increase = 2
        elif choice_function == "pick-two-lose-one":
            self.next_param_func = pick_two_lose_one
            self._numb_func_increase = 1
        else:
            raise ValueError(f"Choice parameter {choice_function} was not recognized.")
        self.num_s = 1
        self.num_p = 0
        self.err_arr = []
        self.fitting_obj = fitting_obj

    @property
    def numb_func_increase(self):
        r"""Return number of basis-functions to add in each iteration."""
        return self._numb_func_increase

    @property
    def density(self):
        r"""Return density that is fitted to."""
        return self.fitting_obj.density

    @property
    def grid(self):
        r"""Return grid class object."""
        return self.fitting_obj.grid

    @property
    def model(self):
        r"""Return model class."""
        return self.fitting_obj.model

    @property
    def measure(self):
        r"""Return measure class."""
        return self.fitting_obj.measure

    @property
    def integral_dens(self):
        r"""Return the integral of the density."""
        return self.fitting_obj.integral_dens

    @abstractmethod
    def get_best_one_function_solution(self):
        r"""Return the solution of the model parameters for one basis function."""
        # Usually the model with single basis function can be solved analytically.
        raise NotImplementedError()

    @abstractmethod
    def get_optimization_routine(self, params, *args, local=False):
        r"""Run the optimization for fitting the Gaussian model to a density."""
        raise NotImplementedError()

    def eval_obj_function(self, params):
        r"""Return evaluation the objective function."""
        model = self.model.evaluate(params[:len(params)//2], params[len(params)//2:])
        integrand = self.measure.evaluate(self.density, model)
        return self.fitting_obj.integrate(integrand)

    def store_errors(self, params):
        r"""Store errors inside the attribute `err_arr`."""
        coeffs, exps = params[:len(params)//2], params[len(params)//2:]
        err = self.fitting_obj.goodness_of_fit(coeffs, exps)
        if len(self.err_arr) == 0:
            self.err_arr = [err]
        else:
            self.err_arr.append(err)

    def _find_best_lparams(self, param_list, num_s_choices, num_p_choices):
        r"""
        Return the best initial guess from a list of potential model parameter choices.

        Parameters
        ----------
        param_list : List[`num_s_choices` + `num_p_choices`]
            List of size `num_s_choices` + `num_p_choices` of each initial guess.
        num_s_choices : int
            Number of guesses for adding extra s-type Gaussians.
        num_p_choices : int
            Number of guesses for adding extra p-type Gaussians.

        Returns
        -------
        (float, List, bool) :
            Returns the value of the cost-function and the best parameters corresponding
            to that value of cost-function.
            True if adding S-type is the most optimal, False otherwise.

        """
        # Initialize the values being returned.
        best_local_value = 1e10
        best_local_param = None
        is_s_optimal = False
        for i in range(0, num_s_choices + num_p_choices):
            # Get the function parameter guess.
            param = param_list[i]

            # Update model for the new number of S-type and P-type functions.
            if i < num_s_choices:
                # Update the number of s-type basis functions for this guess.
                self.model.change_numb_s_and_numb_p(
                    self.num_s + self.numb_func_increase, self.num_p
                )
            else:
                # Update the number of p-type basis functions for this guess.
                self.model.change_numb_s_and_numb_p(
                    self.num_s, self.num_p + self.numb_func_increase
                )

            # Optimize using this initial guess and get cost-function value.
            local_param = self.get_optimization_routine(param, local=True)
            cost_func = self.eval_obj_function(local_param)
            # If it is the best found, then return it.
            if cost_func < best_local_value:
                best_local_value = self.eval_obj_function(local_param)
                best_local_param = local_param
                is_s_optimal = bool(i < num_s_choices)
        return best_local_value, best_local_param, is_s_optimal

    def _split_parameters(self, params):
        r"""Split parameters into the s-type, p-type coefficients and exponents."""
        s_coeffs = params[:self.num_s]
        p_coeffs = params[self.num_s:self.num_s + self.num_p]
        s_exps = params[self.num_s + self.num_p:2 * self.num_s + self.num_p]
        p_exps = params[2 * self.num_s + self.num_p:]
        return s_coeffs, s_exps, p_coeffs, p_exps

    def _print_header(self):
        r"""Print the initial output for the greedy algorithm."""
        # Template for the header.
        print("-" * (10 + 12 + 15 + 15 + 15 + 15 + 15 + 15 + 15 + 17))
        # Format is {identifier:width}, ^ means center it, | means put a bar in it
        template_header = (
            "|{0:^15}|{1:^13}|{2:^13}|{3:^15}|{4:^15}|{5:^16}|{6:^15}|{7:^16}|{8:^16}|"
        )
        # Print the headers
        print(template_header.format(
            "# of Functions", "# of S-type", "# of P-type",
            "Integration", "L1", "L Infinity", "Least-squares",
            "Kullback-Leibler", "Change Objective")
        )
        print("-" * (10 + 12 + 15 + 15 + 15 + 15 + 15 + 15 + 15 + 17))
        # Template for float iteration
        template_iters = (
            "|{0:^15d}|{1:^13d}|{2:^13d}|{3:^15f}|{4:^15e}|{5:^16e}|{6:^15e}|{7:^16e}|{8:^16e}|"
        )
        print(template_iters.format(
            1, self.num_s, self.num_p, *self.err_arr[-1], np.nan
        ))
        return template_iters

    def run(self, factor, d_threshold=1e-8, max_numb_funcs=30, add_extra_choices=None, disp=False):
        r"""
        Add new Gaussians to fit to a density until convergence is achieved.

        Initially, the algorithm solves for the best one-function basis-function.
        Then it generates a group of initial guesses of size `(1 + C)`
        based on the previous optimal fit. It then optimizes each initial guess
        from that group, with a small threshold for convergence.
        It takes the best found initial guess from that set and optimizes it further.
        Then this process repeats for `(1 + 2C)` basis-functions until convergence or
        termination criteria is met.

        Parameters
        ----------
        factor : float
            The factor that is used to generate new initial guess.
        d_threshold : float
            The convergence threshold for the objective function being minimized.
        max_numb_funcs : int
            Maximum number of basis-functions to have.
        add_extra_choices : callable(List, List[List])
            Function that returns extra initial guesses to add.
            Input must be the model parameters and the output should be a
            list of initial guesses that should match attribute `numb_func_increase`.
        disp : bool
            Whether to display the output.

        Returns
        -------
        result : dict
            The optimization results presented as a dictionary containing:

            "coeffs" : ndarray
                The optimized coefficients of Gaussian model.
            "exps" : ndarray
                The optimized exponents of Gaussian model.
            "num_s" : int
                Number of s-type Gaussian functions.
            "num_p" : int
                Number of p-type Gaussian functions.
            "success": bool
                Whether or not the optimization exited successfully.
            "fun" : float
                Objective function at the last iteration.
            "performance" : ndarray
                Values of various performance measures of modeled density at each iteration,
                as computed by `goodness_of_fit()` method.
            "parameters_iteration": List
                List of the optimal parameters of each iteration.
            "exit_information": str
                Information about termination of the greedy algorithm.

        """
        if not isinstance(factor, float):
            raise TypeError(f"Factor {factor} parameter should be a float.")
        if factor <= 0.0:
            raise ValueError(f"Scale {factor} should be positive.")

        # Initialize all the variables
        gparams = self.get_best_one_function_solution()
        self.store_errors(gparams)
        exit_info = None  # String containing information about how it exits.
        numb_funcs = 1  # Number of current functions in model.
        prev_gval, best_gval = np.inf, self.eval_obj_function(gparams)
        params_iter = [gparams]  # Storing the parameters at each iteration.
        numb_redum = 0  # Number of redundancies, termination criteria.
        factor0 = factor  # Scaling parameter that changes.

        # Start the greedy algorithm
        if disp:
            template_iters = self._print_header()

        success = True
        while numb_funcs <= max_numb_funcs - 1 and numb_redum < 5 and \
                np.abs(best_gval - prev_gval) >= d_threshold:
            s_coeffs, s_exps, p_coeffs, p_exps = self._split_parameters(gparams)

            # Get the next list of choices of parameters for S-type and P-type orbitals.
            # Add new S-type and then P-types initial guess.
            choices_parameters_s = self.next_param_func(factor, s_coeffs, s_exps)
            choices_parameters_s = [
                np.hstack((x[:self.num_s + self.numb_func_increase], p_coeffs,
                           x[self.num_s + self.numb_func_increase:], p_exps))
                for x in choices_parameters_s
            ]
            if self.num_p == 0:
                # Get Three Guesses one from 0 to 1, 1 to 10 and 10 to 100.
                choices_parameters_p = [
                    np.random.random((2 * self.numb_func_increase,)),
                    np.random.random((2 * self.numb_func_increase,)) * 9 + 1,
                    np.random.random((2 * self.numb_func_increase,)) * 90 + 10,
                ]
            else:
                choices_parameters_p = self.next_param_func(factor, p_coeffs, p_exps)
            choices_parameters_p = [
                np.hstack((s_coeffs, x[:self.num_p + self.numb_func_increase],
                           s_exps, x[self.num_p + self.numb_func_increase:]))
                for x in choices_parameters_p
            ]
            num_s_choices = len(choices_parameters_s)
            num_p_choices = len(choices_parameters_p)
            total_choices = choices_parameters_s + choices_parameters_p
            if callable(add_extra_choices):
                # When you want extra choices to be added.
                total_choices.append(add_extra_choices(gparams))

            # Run fast, quick optimization and find the best parameter out of the choices.
            _, best_lparam, is_s_optimal = self._find_best_lparams(
                total_choices, num_s_choices, num_p_choices
            )

            # Update model for the new number of S-type and P-type functions.
            if is_s_optimal:
                # If adding s-type was optimal then increase number of s-type.
                self.num_s += self.numb_func_increase
            else:
                self.num_p += self.numb_func_increase
            self.model.change_numb_s_and_numb_p(self.num_s, self.num_p)
            numb_funcs += self.numb_func_increase

            # Optimize further the best local answer and get new objective func
            opt_lparam = self.get_optimization_routine(best_lparam, local=False)
            opt_lvalue = self.eval_obj_function(opt_lparam)

            # Check if redundancies were found in the coefficients and exponents, remove them,
            #   change the factor and try again.
            s_coeffs, s_exps, p_coeffs, p_exps = self._split_parameters(opt_lparam)
            s_coeffs_new, _ = remove_redundancies(s_coeffs, s_exps)
            p_coeffs_new, _ = remove_redundancies(p_coeffs, p_exps)
            found_s_redundances = self.num_s != len(s_coeffs_new)
            found_p_redundancies = self.num_p != len(p_coeffs_new)
            if found_s_redundances or found_p_redundancies or opt_lvalue > best_gval:
                # Move back one function and try a different factor to generate
                #    better initial guesses.
                numb_redum += 1  # Update the termination criteria
                factor *= 2      # Increase the factor that changes the kinds of potential choices.
                if is_s_optimal:
                    self.num_s -= self.numb_func_increase
                else:
                    self.num_p -= self.numb_func_increase
                self.model.change_numb_s_and_numb_p(self.num_s, self.num_p)
                numb_funcs -= self.numb_func_increase
            elif opt_lvalue <= best_gval:
                # Take the best local choice and optimization even further.
                gparams = opt_lparam  # Store for next iteration.
                self.store_errors(gparams)
                params_iter.append(gparams)  # Store parameters.
                prev_gval, best_gval = best_gval, opt_lvalue
                numb_redum = 0    # Reset the number of redundancies.
                factor = factor0  # Reset Original Factor.

            if disp:
                print(template_iters.format(
                    numb_funcs, self.num_s, self.num_p, *self.err_arr[-1],
                    np.abs(best_gval - prev_gval)
                ))

        if numb_funcs == max_numb_funcs - 1 or numb_redum == 5:
            success = False
            if numb_redum == 5:
                exit_info = f"Next Iteration Did Not Find The Best Choice at" \
                            f" number of s-type {self.num_s} and p-type {self.num_p}"

        if exit_info is None:
            exit_info = self._final_exit_info(
                numb_funcs, max_numb_funcs, best_gval, prev_gval, d_threshold
            )

        # Get the coefficients and exponents of the most recent parameters stored.
        coeffs_s, exps_s, coeffs_p, exps_p = self._split_parameters(params_iter[-1])
        obj_func = self.eval_obj_function(params_iter[-1])
        coeffs = np.hstack((coeffs_s, coeffs_p))
        expons = np.hstack((exps_s, exps_p))
        if disp:
            print("-" * (10 + 12 + 15 + 15 + 15 + 15 + 15 + 15 + 15 + 17))
            print(f"Successful?: {success}")
            print(f"Number of s-type: {self.num_s}")
            print(f"Number of p-type: {self.num_p}")
            print("Termination Information: " + exit_info)
            print(f"Objective Function: {obj_func}")
            print(f"Coefficients {coeffs}")
            print(f"Exponents {expons}")

        results = {"coeffs": coeffs,
                   "exps": expons,
                   "num_s": self.num_s,
                   "num_p": self.num_p,
                   "fun": obj_func,
                   "success": success,
                   "parameters_iteration": params_iter,
                   "performance": np.array(self.err_arr),
                   "exit_information": exit_info}
        return results

    @staticmethod
    def _final_exit_info(num_func, max_func, best_val, prev_gval, d_threshold):
        r"""Return string that holds how the greedy algorithm teminated."""
        exit_info = None
        if not num_func < max_func - 1:
            exit_info = "Max number of functions reached,  " + str(num_func)
        elif np.abs(best_val - prev_gval) < d_threshold:
            exit_info = "Cost function is less than some epsilon: " + \
                             str(best_val - prev_gval) + " <= " + str(d_threshold)
        if exit_info is None:
            raise RuntimeError(f"Exit information {exit_info} should not be None. "
                               f"There is an error in the termination of the algorithm.")
        return exit_info


class GreedyLeastSquares(GreedyStrategy):
    r"""Optimize Least-Squares using Greedy and ScipyFit methods."""

    def __init__(
        self, grid, density, choice="pick-one", local_tol=1e-5, global_tol=1e-8,
        method="SLSQP", normalize=False, integral_dens=None, with_constraint=False,
        l_maxiter=1000, g_maxiter=2500, spherical=False,
    ):
        r"""
        Construct the GreedyLestSquares object.

        Parameters
        ----------
        grid : (_BaseRadialGrid, CubicGrid)
            Grid class that contains the grid points and integration methods on them.
        density : ndarray
            The true density evaluated on the grid points.
        choice : str, optional
            Determines how the next set of basis-functions are chosen.
            Can be either:

            * `pick-one` : add a new basis-function by taking the average between every two
              s-type or p-type exponents.
            * `pick-two` : add two new basis-functions by recursion of "pick-one" over each
              guess from `pick-one`.
            * `pick-two-lose-one` : add new basis-function by iterating though each guess
              in `pick-two` and removing one basis-function, generating a new guess each time.
        local_tol : float, optional
            The tolerance for convergence of scipy.optimize method for optimizing each local guess.
            Should be larger than `global_tol`.
        global_tol : float, optional
            The tolerance for convergence of scipy.optimize method for further refining/optimizing
            the best local guess found out of all choices.  Should be smaller than
            `local_tol`.
        method : str, optional
            The method used for optimizing parameters. Default is "slsqp".
            See "scipy.optimize.minimize" for options.
        normalize : bool, optional
            Whether to fit with a normalized s-type and p-type Gaussian model.
        integral_dens : float, optional
            If this is provided, then the model is constrained to integrate to this value.
            If not, then the model is constrained to the numerical integration of the
            density. Useful when one knows the actual integration value of the density.
        with_constraint : bool
            If true, then adds the constraint that the integration of the model density must
            be equal to the constraint of true density. The default is True.
        l_maxiter : int, optional
            Maximum number of iterations when optimizing an initial guess in the `scipy.optimize`
            method for optimizing through all initial guesses (local choices).
        g_maxiter : int, optional
            Maximum number of iterations when optimizing an initial guess in the `scipy.optimize`
            method for further optimizing the best local choice (global choice).
        spherical : bool
            Whether to perform spherical integration by adding :math:`4 \pi r^2` term
            to the integrand when calculating the objective function.
            Only used when grid is one-dimensional and positive (radial grid).

        """
        self._local_tol = local_tol
        self._global_tol = global_tol
        self._l_maxiter = l_maxiter
        self._g_maxiter = g_maxiter
        self._with_constraint = with_constraint
        model = AtomicGaussianDensity(grid.points, num_s=1, num_p=0, normalize=normalize)
        gaussian_obj = ScipyFit(grid, density, model, measure=SquaredDifference(), method=method,
                                integral_dens=integral_dens, spherical=spherical)
        super().__init__(gaussian_obj, choice)

    @property
    def local_tol(self):
        r"""Return local tolerance for convergence in `scipy.optimize`."""
        return self._local_tol

    @property
    def global_tol(self):
        r"""Return global tolerance for convergence in `scipy.optimize`."""
        return self._global_tol

    @property
    def l_maxiter(self):
        r"""Return maximum number of iterations in optimization routine for all local choices."""
        return self._l_maxiter

    @property
    def g_maxiter(self):
        r"""Return maximum number of iterations in optimization routine for global choice."""
        return self._g_maxiter

    @property
    def with_constraint(self):
        r"""Return constraint that integral of model should equal integral of density."""
        return self._with_constraint

    def _solve_one_function_weight(self, weight):
        r"""Solve the best one-basis function solution."""
        a = 2.0 * np.sum(weight)
        sum_of_grid_squared = np.sum(weight * np.power(self.grid.points, 2))
        b = 2.0 * sum_of_grid_squared
        sum_ln_electron_density = np.sum(weight * np.log(self.density))
        c = 2.0 * sum_ln_electron_density
        d = b
        e = 2.0 * np.sum(weight * np.power(self.grid.points, 4))
        f = 2.0 * np.sum(weight * np.power(self.grid.points, 2) * np.log(self.density))
        big_a = (b * f - c * e) / (b * d - a * e)
        big_b = (a * f - c * d) / (a * e - b * d)
        coefficient = np.exp(big_a)
        exponent = - big_b
        return np.array([coefficient, exponent])

    def get_best_one_function_solution(self):
        r"""Obtain the best s-type function solution to least-squares using different weights."""
        # Minimizing weighted least squares with three different weights
        weight1 = np.ones(len(self.grid.points))
        weight3 = np.power(self.density, 2.)
        p1 = self._solve_one_function_weight(weight1)
        cost_func1 = self.eval_obj_function(p1)

        p2 = self._solve_one_function_weight(self.density)
        cost_func2 = self.eval_obj_function(p2)

        p3 = self._solve_one_function_weight(weight3)
        cost_func3 = self.eval_obj_function(p3)

        p_min = min(
            [(cost_func1, p1), (cost_func2, p2), (cost_func3, p3)], key=lambda t: t[0]
        )
        return p_min[1]

    def _create_cofactor_matrix(self, exponents):
        r"""Create cofactor matrix for solving nnls."""
        exponents_s = exponents[:self.model.num_s]
        grid_squared_col = np.power(self.grid.points, 2.0).reshape((len(self.grid.points), 1))
        exponential = np.exp(-exponents_s * grid_squared_col)

        if self.model.num_p != 0:
            exponents_p = exponents[self.model.num_s:]
            exponential = np.hstack(
                (exponential, grid_squared_col * np.exp(-exponents_p * grid_squared_col))
            )
        return exponential

    @staticmethod
    def optimize_using_nnls(true_dens, cofactor_matrix):
        r"""Solve for the coefficients using non-linear least squares."""
        b_vector = np.copy(true_dens)
        b_vector = np.ravel(b_vector)
        row_nnls_coefficients = nnls(cofactor_matrix, b_vector)
        return row_nnls_coefficients[0]

    # pylint: disable=arguments-differ
    def get_optimization_routine(self, params, local=False):
        r"""Optimize least-squares using nnls and scipy.optimize from ScipyFit."""
        # First solves the optimal coefficients (while exponents are fixed) using NNLS.
        # Then it optimizes both coefficients and exponents using scipy.optimize.
        exps = params[len(params)//2:]
        cofac_matrix = self._create_cofactor_matrix(exps)
        coeffs = self.optimize_using_nnls(self.density, cofac_matrix)
        if local:
            results = self.fitting_obj.run(
                coeffs, exps, tol=self.local_tol, maxiter=self.l_maxiter,
                with_constraint=self.with_constraint
            )
        else:
            results = self.fitting_obj.run(
                coeffs, exps, tol=self.global_tol, maxiter=self.g_maxiter,
                with_constraint=self.with_constraint
            )
        return np.hstack((results["coeffs"], results["exps"]))


class GreedyKLFPI(GreedyStrategy):
    r"""Optimize Kullback-Leibler using the Greedy method and fixed point iteration method."""

    def __init__(
        self, grid, density, choice="pick-one", g_eps_coeff=1e-4, g_eps_exp=1e-5,
            g_eps_obj=1e-10, l_eps_coeff=1e-2, l_eps_exp=1e-3, l_eps_obj=1e-8,
            l_maxiter=1000, g_maxiter=2500, mask_value=1e-12, integral_dens=None,
            spherical=False,
    ):
        r"""
        Construct the GreedyKLFPI object.

        Parameters
        ----------
        grid : (_BaseRadialGrid, CubicGrid)
            Grid class that contains the grid points and integration methods on them.
        density : ndarray
            The true density evaluated on the grid points.
        choice : str, optional
            Determines how the next set of basis-functions are chosen.
            Can be either:

            * `pick-one` : add a new basis-function by taking the average between every two
              s-type or p-type exponents.
            * `pick-two` : add two new basis-functions by recursion of "pick-one" over each
              guess from `pick-one`.
            * `pick-two-lose-one` : add new basis-function by iterating though each guess
              in `pick-two` and removing one basis-function, generating a new guess each time.
        g_eps_coeff : float, optional
            The tolerance for convergence of coefficients in KL-FPI method for further
            refining and optimizing the best found local guess.
        g_eps_exp : float, optional
            The tolerance for convergence of exponents in KL-FPI method for further
            refining/optimizing the best local guess found out of all choices.
        g_eps_obj : float, optional
            The tolerance for convergence of objective function in KL-FPI method for
            further refining/optimizing the best local guess found out of all choices.
        l_eps_coeff : float, optional
            The tolerance for convergence of coefficients in KL-FPI method for optimizing
            each local initial guess. Should be larger than `g_eps_coeff`.
        l_eps_exp : float, optional
            The tolerance for convergence of exponents in KL-FPI method for optimizing
            each local initial guess. Should be larger than `g_eps_exp`.
        l_eps_obj : float, optional
            The tolerance for convergence of objective function in KL-FPI method for optimizing
            each local initial guess. Should be larger than `g_eps_obj`.
        l_maxiter : int, optional
            Maximum number of iterations when optimizing an initial guess in the KL-FPI method.
            for optimizing through all initial guesses (local choices).
        g_maxiter : int, optional
            Maximum number of iterations when optimizing an initial guess in the KL-FPI method
            for further optimizing the best local choice (global choice).
        mask_value : str, optional
            The method used for optimizing parameters. Default is "slsqp".
            See "scipy.optimize.minimize" for options.
        integral_dens : float, optional
            If this is provided, then the model is constrained to integrate to this value.
            If not, then the model is constrained to the numerical integration of the
            density. Useful when one knows the actual integration value of the density.
        spherical : bool
            Whether to perform spherical integration by adding :math:`4 \pi r^2` term
            to the integrand when calculating objective function.
            Only used when grid is one-dimensional and positive (radial grid).

        """
        # Algorithm parameters for KL-FPI (KLDivergenceFPI) method.
        self._l_threshold_coeff = l_eps_coeff
        self._l_threshold_exp = l_eps_exp
        self._l_threshold_obj = l_eps_obj
        self._g_threshold_coeff = g_eps_coeff
        self._g_threshold_exp = g_eps_exp
        self._g_threshold_obj = g_eps_obj
        self._l_maxiter = l_maxiter
        self._g_maxiter = g_maxiter
        # Model that is fitted to.
        model = AtomicGaussianDensity(grid.points, num_s=1, num_p=0, normalize=True)
        fpi_obj = KLDivergenceFPI(grid, density, model, mask_value, integral_dens, spherical)
        super().__init__(fpi_obj, choice)

    @property
    def l_threshold_coeff(self):
        r"""Return local threshold for KL-FPI method for convergence of coefficients."""
        return self._l_threshold_coeff

    @property
    def l_threshold_exp(self):
        r"""Return local threshold for KL-FPI method for convergence of exponents."""
        return self._l_threshold_exp

    @property
    def l_threshold_obj(self):
        r"""Return local threshold for KL-FPI method for convergence of objective function."""
        return self._l_threshold_obj

    @property
    def g_threshold_coeff(self):
        r"""Return global threshold for KL-FPI method for convergence of coefficients."""
        return self._g_threshold_coeff

    @property
    def g_threshold_exp(self):
        r"""Return global threshold for KL-FPI method for convergence of exponents."""
        return self._g_threshold_exp

    @property
    def g_threshold_obj(self):
        r"""Return global threshold for KL-FPI method for convergence of objective function."""
        return self._g_threshold_obj

    @property
    def l_maxiter(self):
        r"""Return maximum number of iterations in optimization routine for all local choices."""
        return self._l_maxiter

    @property
    def g_maxiter(self):
        r"""Return maximum number of iterations in optimization routine for global choice."""
        return self._g_maxiter

    def get_best_one_function_solution(self):
        r"""Obtain the best one s-type function to Kullback-Leibler."""
        denom = self.grid.integrate(self.density * np.power(self.grid.points ** 2, 2.))
        exps = 3. * self.integral_dens / (2. * 4. * np.pi * denom)
        return np.array([self.integral_dens, exps])

    # pylint: disable=arguments-differ
    def get_optimization_routine(self, params, local=False):
        r"""Optimize KL using KL-FPI method."""
        coeffs, exps = params[:len(params)//2], params[len(params)//2:]
        if local:
            result = self.fitting_obj.run(
                coeffs, exps, opt_coeffs=True, opt_expons=True,
                maxiter=self.l_maxiter, c_threshold=self.l_threshold_coeff,
                e_threshold=self.l_threshold_exp, d_threshold=self.l_threshold_obj,
                disp=False
            )
            return np.hstack((result["coeffs"], result["exps"]))
        result = self.fitting_obj.run(
            coeffs, exps, opt_coeffs=True, opt_expons=True, maxiter=self.g_maxiter,
            c_threshold=self.g_threshold_coeff, e_threshold=self.g_threshold_exp,
            d_threshold=self.g_threshold_obj, disp=False
        )
        return np.hstack((result["coeffs"], result["exps"]))



# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
r"""Test bfit.grid module."""

import numpy as np
from numpy.testing import assert_almost_equal, assert_raises

from bfit.grid import _BaseRadialGrid, ClenshawRadialGrid, CubicGrid, UniformRadialGrid


def test_raises_base():
    r"""Test _BaseRadialGrid raises errors."""
    # check array
    assert_raises(TypeError, _BaseRadialGrid, 5.)
    assert_raises(TypeError, _BaseRadialGrid, [1., 2., 3.])
    assert_raises(TypeError, _BaseRadialGrid, (1., 2., 3.))
    # check 1D array
    assert_raises(TypeError, _BaseRadialGrid, np.array([[5.]]))
    assert_raises(TypeError, _BaseRadialGrid, np.array([[5., 5.5], [1.3, 2.5]]))


def test_integration_base():
    r"""Test integration of _BaseRadialGrid."""
    # integrate a triangle
    grid = _BaseRadialGrid(np.arange(0., 2., 0.000001))
    value = grid.integrate(grid.points)
    assert_almost_equal(value, 2. * 2. / 2, decimal=5)
    # integrate a square
    value = grid.integrate(2. * np.ones(len(grid)))
    assert_almost_equal(value, 2. * 2., decimal=5)


def test_raises_integration():
    r"""Test integration over BaseRadialGrid returns an error if dimension aren't specified."""
    grid = _BaseRadialGrid(np.arange(0., 2., 0.000001))
    # Obtain an array with different dimension than the "grid.points".
    arr = np.arange(0., 2., 0.1)
    assert_raises(ValueError, grid.integrate, arr)


def test_raises_clenshaw():
    r"""Test ClenshawRadialGrid raises error."""
    assert_raises(TypeError, ClenshawRadialGrid, 10.1, 1, 1, [])
    assert_raises(TypeError, ClenshawRadialGrid, -10, 1, 1, [])
    assert_raises(TypeError, ClenshawRadialGrid, 1, 2.2, 1, [])
    assert_raises(TypeError, ClenshawRadialGrid, 1, -2, 1, [])
    assert_raises(TypeError, ClenshawRadialGrid, 1, 1, 1.1, [])
    assert_raises(TypeError, ClenshawRadialGrid, 1, 1, -2, [])
    assert_raises(TypeError, ClenshawRadialGrid, 1, 1, 1, "not list")
    assert_raises(TypeError, ClenshawRadialGrid, 1, 1, 1, [], "not bool")


def test_raises_uniform():
    r"""Test UniformRadialGrid raises error."""
    assert_raises(TypeError, UniformRadialGrid, 10.1, 1, 1)
    assert_raises(TypeError, UniformRadialGrid, -10, 1, 1)
    assert_raises(TypeError, UniformRadialGrid, 10, "not numb", 1)
    assert_raises(TypeError, UniformRadialGrid, 10, 1, "not numb")
    assert_raises(ValueError, UniformRadialGrid, 1, 2.2, 1)


def test_points_clenshaw():
    r"""Test points in the ClenshawRadialGrid."""
    grid = ClenshawRadialGrid(10, 10, 20, [1000])
    # check core points
    core = [(1. - np.cos(np.pi * x / 20)) / 20. for x in range(0, 10)]
    assert_almost_equal(core, grid._get_points(10, "core"), decimal=8)
    # check diffuse points
    diff = [25. * (1. - np.cos(np.pi * x / 40.)) for x in range(1, 21)]
    assert_almost_equal(diff, grid._get_points(20, "diffuse"), decimal=8)
    # check points
    assert_almost_equal(sorted(core + diff + [1000]), grid.points, decimal=8)
    # check assertion
    assert_raises(ValueError, grid._get_points, 10, "not core or diffuse")
    # test unique number of points
    assert len(np.unique(grid.points)) == len(grid.points)


def test_points_clenshaw_not_including_origin():
    r"""Test points in the ClenshawRadialGrid without the origin."""
    grid = ClenshawRadialGrid(10, 11, 20, [1000], include_origin=False)
    # check core points
    core = [(1. - np.cos(np.pi * x / 22.0)) / 20. for x in range(1, 12)]
    core_actual = grid._get_points(11, "core", include_origin=False)
    assert_almost_equal(core, core_actual, decimal=8)
    assert len(core_actual) == 11
    # check diffuse points
    diff = [25. * (1. - np.cos(np.pi * x / 40.)) for x in range(1, 21)]
    valence_actual = grid._get_points(20, "diffuse", include_origin=False)
    assert_almost_equal(diff, valence_actual, decimal=8)
    assert len(valence_actual) == 20
    # check points
    assert_almost_equal(sorted(core + diff + [1000]), grid.points, decimal=8)
    # check assertion
    assert_raises(ValueError, grid._get_points, 10, "not core or diffuse")
    # test unique number of points
    assert len(np.unique(grid.points)) == len(grid.points)


def test_points_uniform():
    r"""Test points in UniformRadialGrid."""
    # check 1 point
    grid = UniformRadialGrid(1, 0, 1)
    assert_almost_equal(grid.points, np.array([0.]), decimal=8)
    # check 2 point
    grid = UniformRadialGrid(2, 0, 1)
    assert_almost_equal(grid.points, np.array([0, 1.]), decimal=8)
    # check 5 points
    grid = UniformRadialGrid(5, 0, 33)
    assert_almost_equal(grid.points, np.arange(0, 33.1, step=33/4.), decimal=8)
    # check 10 points
    grid = UniformRadialGrid(10, 0, 100)
    assert_almost_equal(grid.points, np.arange(0, 100.1, step=100/9.), decimal=8)


def test_integration_uniform_gaussian():
    r"""Test integration of UniformRadialGrid of different Gaussians."""
    grid = UniformRadialGrid(100, 0.0, 15.0)
    pnts = grid.points
    spherical_comp = 4.0 * np.pi * pnts**2.0
    # integrate s-type gaussian functions
    value = grid.integrate(np.exp(-pnts**2) * 4.0 * np.pi * pnts**2.0)
    assert_almost_equal(value, np.pi**1.5, decimal=6)
    value = grid.integrate(1.23 * np.exp(-0.5 * pnts**2) * 4.0 * np.pi * pnts**2.0)
    assert_almost_equal(value, 1.23 * (np.pi / 0.5)**1.5, decimal=6)
    value += grid.integrate(0.91 * np.exp(-0.1 * pnts**2) * 4.0 * np.pi * pnts**2.0)
    assert_almost_equal(value, 1.23 * (np.pi / 0.5)**1.5 + 0.91 * (np.pi / 0.1)**1.5, decimal=6)
    # integrate p-type gaussian functions
    value = grid.integrate(grid.points**2 * np.exp(-2.5 * pnts**2) * spherical_comp)
    assert_almost_equal(value, 1.5 * np.pi**1.5 / 2.5**2.5, decimal=6)
    value = grid.integrate(1.2 * pnts**2 * np.exp(-pnts**2) * spherical_comp)
    assert_almost_equal(value, 1.2 * 1.5 * np.pi**1.5, decimal=6)
    value = grid.integrate(2.5 * pnts**2 * np.exp(-1.1 * pnts**2) * spherical_comp)
    assert_almost_equal(value, 2.5 * 1.5 * np.pi**1.5 / 1.1**2.5, decimal=6)
    value = grid.integrate(
        2. * 2.75**2.5 * pnts**2 * np.exp(-2.75 * pnts**2) / (3. * np.pi**1.5) * spherical_comp
    )
    assert_almost_equal(value, 1.0, decimal=6)
    # integrate s-type + p-type gaussians
    value = grid.integrate(
        (np.exp(-4.01 * pnts**2) + pnts**2 * np.exp(-1.25 * pnts**2)) * spherical_comp
    )
    assert_almost_equal(value, (np.pi / 4.01)**1.5 + 1.5 * (np.pi**1.5 / 1.25**2.5), decimal=6)
    value = (0.5 / np.pi)**1.5 * np.exp(-0.5 * pnts**2) + 3.62 * pnts**2 * np.exp(-0.85 * pnts**2)
    value = grid.integrate(value * spherical_comp)
    assert_almost_equal(value, 1.0 + 3.62 * 1.5 * (np.pi**1.5 / 0.85**2.5), decimal=6)


def test_integration_uniform_gaussian_on_x_axis():
    r"""Test integration of UniformRadialGrid of different Gaussians."""
    # integrating from -15 to 15, which is twice as much as integrating from 0 to 15
    grid = UniformRadialGrid(300, -15., 15.)
    pnts = grid.points
    spherical = 4.0 * np.pi * pnts**2.0
    # integrate s-type gaussian functions
    value = grid.integrate(np.exp(-pnts**2) * spherical)
    assert_almost_equal(value, 2 * np.pi**1.5, decimal=6)
    value = grid.integrate(1.23 * np.exp(-0.5 * pnts**2) * spherical)
    assert_almost_equal(value, 2 * 1.23 * (np.pi / 0.5)**1.5, decimal=6)
    value += grid.integrate(0.91 * np.exp(-0.1 * pnts**2) * spherical)
    expected_value = 2 * (1.23 * (np.pi / 0.5)**1.5 + 0.91 * (np.pi / 0.1)**1.5)
    assert_almost_equal(value, expected_value, decimal=6)
    # integrate p-type gaussian functions
    value = grid.integrate(pnts**2 * np.exp(-2.5 * pnts**2) * spherical)
    assert_almost_equal(value, 2 * 1.5 * np.pi**1.5 / 2.5**2.5, decimal=6)
    value = grid.integrate(1.2 * pnts**2 * np.exp(-pnts**2) * spherical)
    assert_almost_equal(value, 2 * 1.2 * 1.5 * np.pi**1.5, decimal=6)
    value = grid.integrate(2.5 * pnts**2 * np.exp(-1.1 * pnts**2) * spherical)
    assert_almost_equal(value, 2 * 2.5 * 1.5 * np.pi**1.5 / 1.1**2.5, decimal=6)
    value = grid.integrate(
        2. * 2.75**2.5 * pnts**2 * np.exp(-2.75 * pnts**2) / (3. * np.pi**1.5) * spherical
    )
    assert_almost_equal(value, 2 * 1.0, decimal=6)
    # integrate s-type + p-type gaussians
    value = grid.integrate(
        (np.exp(-4.01 * pnts**2) + pnts**2 * np.exp(-1.25 * pnts**2)) * spherical
    )
    expected_value = 2 * ((np.pi / 4.01)**1.5 + 1.5 * (np.pi**1.5 / 1.25**2.5))
    assert_almost_equal(value, expected_value, decimal=6)
    value = (0.5 / np.pi)**1.5 * np.exp(-0.5 * pnts**2) + 3.62 * pnts**2 * np.exp(-0.85 * pnts**2)
    value = grid.integrate(value * spherical)
    assert_almost_equal(value, 2 * (1.0 + 3.62 * 1.5 * (np.pi**1.5 / 0.85**2.5)), decimal=6)


def test_integration_uniform_gaussian_shifted():
    r"""Test integration of a shifted Gaussian using UniformRadialGrid."""
    # spherical=False is used but r^2 is added to the integrand
    grid = UniformRadialGrid(300, 0.0, 30.0)
    pnts = grid.points - 15.
    # integrate s-type gaussian functions
    value = 4 * np.pi * grid.integrate(pnts**2 * np.exp(-pnts**2))
    assert_almost_equal(value, 2 * np.pi**1.5, decimal=6)
    value = 4 * np.pi * grid.integrate(1.23 * pnts**2 * np.exp(-0.5 * pnts**2))
    assert_almost_equal(value, 2 * 1.23 * (np.pi / 0.5)**1.5, decimal=6)
    value += 4 * np.pi * grid.integrate(0.91 * pnts**2 * np.exp(-0.1 * pnts**2))
    expected_value = 2 * (1.23 * (np.pi / 0.5)**1.5 + 0.91 * (np.pi / 0.1)**1.5)
    assert_almost_equal(value, expected_value, decimal=6)
    # integrate p-type gaussian functions
    value = 4 * np.pi * grid.integrate(pnts**4 * np.exp(-2.5 * pnts**2))
    assert_almost_equal(value, 2 * 1.5 * np.pi**1.5 / 2.5**2.5, decimal=6)
    value = 4 * np.pi * grid.integrate(1.2 * pnts**4 * np.exp(-pnts**2))
    assert_almost_equal(value, 2 * 1.2 * 1.5 * np.pi**1.5, decimal=6)
    value = 4 * np.pi * grid.integrate(2.5 * pnts**4 * np.exp(-1.1 * pnts**2))
    assert_almost_equal(value, 2 * 2.5 * 1.5 * np.pi**1.5 / 1.1**2.5, decimal=6)
    value = grid.integrate(2. * 2.75**2.5 * pnts**4 * np.exp(-2.75 * pnts**2) / (3. * np.pi**1.5))
    assert_almost_equal(4 * np.pi * value, 2 * 1.0, decimal=6)
    # integrate s-type + p-type gaussians
    value = grid.integrate(pnts**2 * np.exp(-4.01 * pnts**2) + pnts**4 * np.exp(-1.25 * pnts**2))
    value *= 4 * np.pi
    expected_value = 2 * ((np.pi / 4.01)**1.5 + 1.5 * (np.pi**1.5 / 1.25**2.5))
    assert_almost_equal(value, expected_value, decimal=6)
    value = (0.5 / np.pi)**1.5 * pnts**2 * np.exp(-0.5 * pnts**2)
    value += 3.62 * pnts**4 * np.exp(-0.85 * pnts**2)
    value = 4 * np.pi * grid.integrate(value)
    assert_almost_equal(value, 2 * (1.0 + 3.62 * 1.5 * (np.pi**1.5 / 0.85**2.5)), decimal=6)


def test_integration_clenshaw_gaussian():
    r"""Test integration of ClenshawRadialGrid of different Gaussians."""
    # clenshaw grid & points
    grid = ClenshawRadialGrid(10, 10000, 10000)
    pnts = grid.points
    spherical = 4.0 * np.pi * pnts**2.0
    # integrate s-type gaussian functions
    value = grid.integrate(np.exp(-pnts**2) * spherical)
    assert_almost_equal(value, np.pi**1.5, decimal=6)
    value = grid.integrate(1.23 * np.exp(-0.5 * pnts**2) * spherical)
    assert_almost_equal(value, 1.23 * (np.pi / 0.5)**1.5, decimal=6)
    value += grid.integrate(0.91 * np.exp(-0.1 * pnts**2) * spherical)
    assert_almost_equal(value, 1.23 * (np.pi / 0.5)**1.5 + 0.91 * (np.pi / 0.1)**1.5, decimal=6)
    # integrate p-type gaussian functions
    value = grid.integrate(grid.points**2 * np.exp(-2.5 * pnts**2) * spherical)
    assert_almost_equal(value, 1.5 * np.pi**1.5 / 2.5**2.5, decimal=6)
    value = grid.integrate(1.2 * pnts**2 * np.exp(-pnts**2) * spherical)
    assert_almost_equal(value, 1.2 * 1.5 * np.pi**1.5, decimal=6)
    value = grid.integrate(2.5 * pnts**2 * np.exp(-1.1 * pnts**2) * spherical)
    assert_almost_equal(value, 2.5 * 1.5 * np.pi**1.5 / 1.1**2.5, decimal=6)
    value = grid.integrate(
        2. * 2.75**2.5 * pnts**2 * np.exp(-2.75 * pnts**2) / (3. * np.pi**1.5) * spherical
    )
    assert_almost_equal(value, 1.0, decimal=6)
    # integrate s-type + p-type gaussians
    value = grid.integrate(
        (np.exp(-4.01 * pnts**2) + pnts**2 * np.exp(-1.25 * pnts**2)) * spherical
    )
    assert_almost_equal(value, (np.pi / 4.01)**1.5 + 1.5 * (np.pi**1.5 / 1.25**2.5), decimal=6)
    value = (0.5 / np.pi)**1.5 * np.exp(-0.5 * pnts**2) + 3.62 * pnts**2 * np.exp(-0.85 * pnts**2)
    value = grid.integrate(value * spherical)
    assert_almost_equal(value, 1.0 + 3.62 * 1.5 * (np.pi**1.5 / 0.85**2.5), decimal=6)


def test_raises_cubic():
    r"""Test CubicGrid raises error."""
    assert_raises(AttributeError, CubicGrid, "blah", 2., 3.)
    assert_raises(AttributeError, CubicGrid, -1., "blah", 3.)
    assert_raises(AttributeError, CubicGrid, -1., 2., "blah")
    assert_raises(AttributeError, CubicGrid, -1., 2., -5.)
    assert_raises(AttributeError, CubicGrid, 5., -5., 3.)


def test_integration_cubic():
    r"""Test integration of a cubic grid."""
    axes = np.array([[0.01, 0.0, 0.0], [0.0, 0.01, 0.0], [0.0, 0.0, 0.01]])
    grid = CubicGrid(np.zeros(3), axes, (25, 25, 25))
    # integrate constant value of 1.
    value = grid.integrate(np.ones(len(grid)))
    assert_almost_equal(value, 0.25**3, decimal=3)
    # integrate constant value of 2.
    value = grid.integrate(2 * np.ones(len(grid)))
    assert_almost_equal(value, 2 * 0.25**3, decimal=3)
    # return error if arr is not the same length.
    assert_raises(ValueError, grid.integrate, np.arange(0., 0.25, 0.1))


def test_integration_cubic_gaussian():
    r"""Test integration of a Gaussian using CubicGrid."""
    axes = np.array([[0.25, 0.0, 0.0], [0.0, 0.25, 0.0], [0.0, 0.0, 0.25]])
    grid = CubicGrid(np.array([-15.0, -15.0, -15.0]), axes, (120, 120, 120))
    dist = np.linalg.norm(grid.points, axis=1)
    # integrate s-type gaussian functions
    value = grid.integrate(np.exp(-dist**2))
    assert_almost_equal(value, np.pi**1.5, decimal=6)
    value = grid.integrate(1.23 * np.exp(-0.5 * dist**2))
    assert_almost_equal(value, 1.23 * (np.pi / 0.5)**1.5, decimal=6)
    value += grid.integrate(0.91 * np.exp(-0.1 * dist**2))
    assert_almost_equal(value, 1.23 * (np.pi / 0.5)**1.5 + 0.91 * (np.pi / 0.1)**1.5, decimal=6)
    # integrate p-type gaussian functions
    value = grid.integrate(dist**2 * np.exp(-2.5 * dist**2))
    assert_almost_equal(value, 1.5 * np.pi**1.5 / 2.5**2.5, decimal=6)
    value = grid.integrate(1.2 * dist**2 * np.exp(-dist**2))
    assert_almost_equal(value, 1.2 * 1.5 * np.pi**1.5, decimal=6)
    value = grid.integrate(2.5 * dist**2 * np.exp(-1.1 * dist**2))
    assert_almost_equal(value, 2.5 * 1.5 * np.pi**1.5 / 1.1**2.5, decimal=6)
    value = grid.integrate(2. * 2.75**2.5 * dist**2 * np.exp(-2.75 * dist**2) / (3. * np.pi**1.5))
    assert_almost_equal(value, 1.0, decimal=6)
    # integrate s-type + p-type gaussian functions
    value = grid.integrate(np.exp(-4.01 * dist**2) + dist**2 * np.exp(-1.25 * dist**2))
    assert_almost_equal(value, (np.pi / 4.01)**1.5 + 1.5 * (np.pi**1.5 / 1.25**2.5), decimal=6)
    value = (0.5 / np.pi)**1.5 * np.exp(-0.5 * dist**2) + 3.62 * dist**2 * np.exp(-0.85 * dist**2)
    value = grid.integrate(value)
    assert_almost_equal(value, 1.0 + 3.62 * 1.5 * (np.pi**1.5 / 0.85**2.5), decimal=6)


def test_integration_cubic_gaussian_shifted():
    r"""Test integration of a shifted Gaussian using CubicGrid."""
    axes = np.array([[0.25, 0.0, 0.0], [0.0, 0.25, 0.0], [0.0, 0.0, 0.25]])
    grid = CubicGrid(np.zeros(3), axes, (120, 120, 120))
    # place the function at the center of cubic grid with coordinates of [15, 15, 15]
    dist = np.linalg.norm(grid.points - np.array([15., 15., 15.]), axis=1)
    # integrate s-type gaussian functions
    value = grid.integrate(np.exp(-dist**2))
    assert_almost_equal(value, np.pi**1.5, decimal=6)
    value = grid.integrate(1.23 * np.exp(-0.5 * dist**2))
    assert_almost_equal(value, 1.23 * (np.pi / 0.5)**1.5, decimal=6)
    value += grid.integrate(0.91 * np.exp(-0.1 * dist**2))
    expected_value = 1.23 * (np.pi / 0.5)**1.5 + 0.91 * (np.pi / 0.1)**1.5
    assert_almost_equal(value, expected_value, decimal=6)
    # integrate p-type gaussian functions
    value = grid.integrate(dist**2 * np.exp(-2.5 * dist**2))
    assert_almost_equal(value, 1.5 * np.pi**1.5 / 2.5**2.5, decimal=6)
    value = grid.integrate(1.2 * dist**2 * np.exp(-dist**2))
    assert_almost_equal(value, 1.2 * 1.5 * np.pi**1.5, decimal=6)
    value = grid.integrate(2.5 * dist**2 * np.exp(-1.1 * dist**2))
    assert_almost_equal(value, 2.5 * 1.5 * np.pi**1.5 / 1.1**2.5, decimal=6)
    value = grid.integrate(2. * 2.75**2.5 * dist**2 * np.exp(-2.75 * dist**2) / (3. * np.pi**1.5))
    assert_almost_equal(value, 1.0, decimal=6)
    # integrate s-type + p-type gaussian functions
    value = grid.integrate(np.exp(-4.01 * dist**2) + dist**2 * np.exp(-1.25 * dist**2))
    expected_value = (np.pi / 4.01)**1.5 + 1.5 * (np.pi**1.5 / 1.25**2.5)
    assert_almost_equal(value, expected_value, decimal=6)
    value = (0.5 / np.pi)**1.5 * np.exp(-0.5 * dist**2) + 3.62 * dist**2 * np.exp(-0.85 * dist**2)
    value = grid.integrate(value)
    assert_almost_equal(value, 1.0 + 3.62 * 1.5 * (np.pi**1.5 / 0.85**2.5), decimal=6)



# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
r"""Test bfit.fit module."""

import numpy as np
from numpy.testing import assert_almost_equal, assert_equal, assert_raises

from bfit.fit import KLDivergenceFPI, ScipyFit
from bfit.grid import CubicGrid, UniformRadialGrid
from bfit.measure import KLDivergence, SquaredDifference
from bfit.model import AtomicGaussianDensity, MolecularGaussianDensity


def test_lagrange_multiplier():
    r"""Test lagrange multiplier of KL-SCF."""
    g = UniformRadialGrid(150, 1e-4, 15.0)
    e = np.exp(-g.points)
    kl = KLDivergenceFPI(g, e, None, spherical=False)
    assert_almost_equal(kl.lagrange_multiplier, 1., decimal=8)


def test_goodness_of_fit():
    r"""Test goodness of fit with analytic values."""
    g = UniformRadialGrid(1000, 0.0, 10.0)
    e = np.exp(-g.points)
    m = AtomicGaussianDensity(g.points, num_s=1, num_p=0, normalize=False)
    kl = KLDivergenceFPI(g, e, m, mask_value=0., spherical=True)
    gf = kl.goodness_of_fit(np.array([1.]), np.array([1.]))
    expected = [
        5.56833, 4 * np.pi * 1.60909, 0.128, 4.0 * np.pi * 0.0882922, 4. * np.pi * 17.360
    ]
    assert_almost_equal(expected, gf, decimal=1)


def test_assertion_raises():
    r"""Test assertion raises of all fitting methods."""
    g = UniformRadialGrid(1000, 0.0, 10.0)
    e = np.exp(-g.points)
    m = AtomicGaussianDensity(g.points, num_s=1, num_p=0, normalize=False)
    kl = KLDivergenceFPI(g, e, m, mask_value=0., spherical=True)
    assert_raises(ValueError, kl._update_params, None, None, False, False)
    assert_raises(ValueError, kl.run, np.array([1., 2.]), np.array([1.]))
    assert_raises(ValueError, kl.run, np.array([1.]), np.array([1., 2.]))
    assert_raises(ValueError, kl.run, np.array([1.]), np.array([1.]), False, False)
    assert not kl.run(np.array([1.]), np.array([1.]), maxiter=1)["success"]

    m2 = AtomicGaussianDensity(g.points * 2., num_s=1, num_p=0, normalize=False)
    measure = KLDivergence()
    assert_raises(ValueError, ScipyFit, g, e, m2)
    assert_raises(ValueError, ScipyFit, g, np.append(e, 2.), m)
    assert_raises(ValueError, ScipyFit, g, e, m, measure, "non slsqp")
    assert_raises(TypeError, ScipyFit, g, e, m, "not kl")
    gb = ScipyFit(g, e, m, measure=measure)
    assert_raises(ValueError, gb.run, [], [], False, False)
    assert_raises(ValueError, gb.evaluate_model, [], ("not fixed", 2))

    # Test giving a grid class with no points returns error
    class GridNoPoints:
        def integrate(self):
            pass
    assert_raises(AttributeError, ScipyFit, GridNoPoints(), e, KLDivergence())

    # Test giving grid class with no integrate method returns error.
    class GridNoIntegrate:
        def __init__(self):
            self.points = g.points
    assert_raises(AttributeError, ScipyFit, GridNoIntegrate(), e, KLDivergence())


def test_run_normalized_s_gaussian():
    r"""Test KL-SCF against normalized s-type Gaussian."""
    # density is normalized 1s orbital with exponent=1.0
    g = UniformRadialGrid(150, 0.0, 15.0)
    e = (1. / np.pi)**1.5 * np.exp(-g.points**2.)
    model = AtomicGaussianDensity(g.points, num_s=1, num_p=0, normalize=True)
    kl = KLDivergenceFPI(g, e, model, spherical=True)

    # fit density with initial coeff=1. & expon=1.
    res = kl.run(np.array([1.]), np.array([1.]), True, True, 500, 1.e-4, 1.e-4, 1.e-4)
    # check optimized coeffs & expons
    assert_almost_equal(np.array([1.]), res["coeffs"], decimal=8)
    assert_almost_equal(np.array([1.]), res["exps"], decimal=8)
    # check value of optimized objective function & fitness measure
    assert_almost_equal(0., res["fun"][-1], decimal=8)
    assert_almost_equal(1., res["performance"][-1, 0], decimal=8)
    assert_almost_equal(0., res["performance"][-1, 1:], decimal=8)

    # fit density with initial coeff=0.5 & expon=0.5
    res = kl.run(np.array([0.5]), np.array([0.5]), True, True, 500, 1.e-4, 1.e-4, 1.e-4)
    # check optimized coeffs & expons
    assert_almost_equal(np.array([1.]), res["coeffs"], decimal=8)
    assert_almost_equal(np.array([1.]), res["exps"], decimal=8)
    # check value of optimized objective function & fitness measure
    assert_almost_equal(0., res["fun"][-1], decimal=8)
    assert_almost_equal(1., res["performance"][-1, 0], decimal=8)
    assert_almost_equal(0., res["performance"][-1, 1:], decimal=8)

    # fit density with initial coeff=0.1 & expon=10.
    res = kl.run(np.array([0.1]), np.array([10.]), True, True, 500, 1.e-4, 1.e-4, 1.e-4)
    # check optimized coeffs & expons
    assert_almost_equal(np.array([1.]), res["coeffs"], decimal=8)
    assert_almost_equal(np.array([1.]), res["exps"], decimal=8)
    # check value of optimized objective function
    assert_almost_equal(0., res["fun"][-1], decimal=8)

    # fit density with initial coeff=20. & expon=0.01.
    res = kl.run(np.array([20.]), np.array([0.01]), True, True, 500, 1.e-4, 1.e-4, 1.e-4)
    # check optimized coeffs & expons
    assert_almost_equal(np.array([1.]), res["coeffs"], decimal=8)
    assert_almost_equal(np.array([1.]), res["exps"], decimal=8)
    # check value of optimized objective function
    assert_almost_equal(0., res["fun"][-1], decimal=8)


def test_kl_scf_update_coeffs_2s_gaussian():
    r"""Test KL-SCF method for updating coefficients of two s-type Gaussians."""
    # actual density is a 1s Slater function
    grid = UniformRadialGrid(150, 0.0, 15.0)
    c, e = np.array([5., 2.]), np.array([10., 3.])
    dens = np.exp(-grid.points, dtype=np.float64)
    # model density is a normalized 2s Gaussian basis
    model = AtomicGaussianDensity(grid.points, num_s=2, num_p=0, normalize=True)
    # test updating coeffs
    kl = KLDivergenceFPI(grid, dens, model, spherical=False)
    new_coeffs, new_expons = kl._update_params(c, e, True, False)
    # compute model density
    approx = c[0] * (e[0] / np.pi)**1.5 * np.exp(-e[0] * grid.points**2)
    approx += c[1] * (e[1] / np.pi)**1.5 * np.exp(-e[1] * grid.points**2)
    # compute expected coeffs
    coeffs = c * (e / np.pi)**1.5
    ratio = dens / approx
    coeffs[0] *= grid.integrate(ratio * np.exp(-e[0] * grid.points**2))
    coeffs[1] *= grid.integrate(ratio * np.exp(-e[1] * grid.points**2))
    assert_almost_equal(new_expons, e, decimal=6)
    assert_almost_equal(new_coeffs, coeffs, decimal=6)


def test_kl_scf_update_params_2s_gaussian():
    r"""Test KL-SCF method for updating exponents of two s-type Gaussians."""
    # actual density is a 1s Slater function
    grid = UniformRadialGrid(1000, 0.0, 20.0)
    points = grid.points
    spherical = 4.0 * np.pi * points**2.0
    c, e = np.array([5., 2.], dtype=np.float64), np.array([10., 3.], dtype=np.float64)
    dens = np.exp(-points, dtype=np.float64)

    # model density is a normalized 2s Gaussian basis
    model = AtomicGaussianDensity(points, num_s=2, num_p=0, normalize=True)
    # test updating coeffs
    kl = KLDivergenceFPI(grid, dens, model, spherical=True)
    new_coeffs, new_expons = kl._update_params(c, e, False, True)
    # compute model density
    approx = c[0] * (e[0] / np.pi)**1.5 * np.exp(-e[0] * points**2)
    approx += c[1] * (e[1] / np.pi)**1.5 * np.exp(-e[1] * points**2)
    # compute expected expons
    expons = 1.5 * np.ones(2)
    ratio = np.ma.filled(dens / np.ma.array(approx), 0.)
    expons[0] *= grid.integrate(np.exp(-e[0] * points**2) * ratio * spherical)
    expons[1] *= grid.integrate(np.exp(-e[1] * points**2) * ratio * spherical)
    expons[0] /= grid.integrate((np.exp(-e[0] * points**2) * points**2) * ratio * spherical)
    expons[1] /= grid.integrate((np.exp(-e[1] * points**2) * points**2) * ratio * spherical)
    assert_almost_equal(new_coeffs, c, decimal=6)
    assert_almost_equal(new_expons, expons, decimal=6)


def test_kl_scf_update_params_1s1p_gaussian():
    r"""Test KL-SCF method against one s-type and one p-type Gaussians."""
    # actual density is a 1s Slater function
    grid = UniformRadialGrid(150, 0.0, 15.0)
    points = grid.points
    spherical = 4.0 * np.pi * points**2.0
    c, e = np.array([1., 2.]), np.array([3., 4.])
    dens = np.exp(-grid.points)
    # model density is a normalized 2s Gaussian basis
    model = AtomicGaussianDensity(points, num_s=1, num_p=1, normalize=True)
    # compute model density
    approx = c[0] * (e[0] / np.pi)**1.5 * np.exp(-e[0] * points**2)
    approx += c[1] * 2. * e[1]**2.5 * points**2 * np.exp(-e[1] * points**2) / (3 * np.pi**1.5)
    # check model.evaluate
    assert_almost_equal(approx, model.evaluate(c, e), decimal=6)
    # test updating coeffs
    kl = KLDivergenceFPI(grid, dens, model, mask_value=0., spherical=True)
    new_coeffs, new_expons = kl._update_params(c, e, update_coeffs=True, update_expons=False)
    coeffs = c * np.array([(e[0] / np.pi)**1.5, 2 * e[1]**2.5 / (3 * np.pi**1.5)])
    coeffs[0] *= grid.integrate(dens * np.exp(-e[0] * points**2) / approx * spherical)
    coeffs[1] *= grid.integrate(dens * np.exp(-e[1] * points**2) * points**2 / approx * spherical)
    assert_almost_equal(new_expons, e, decimal=6)
    assert_almost_equal(new_coeffs, coeffs, decimal=6)
    # test updating expons
    new_coeffs, new_expons = kl._update_params(c, e, update_coeffs=False, update_expons=True)
    expons = np.array([1.5, 2.5])
    expons[0] *= grid.integrate(dens * np.exp(-e[0] * points**2) / approx * spherical)
    expons[1] *= grid.integrate(dens * points**2 * np.exp(-e[1] * points**2) / approx * spherical)
    expons[0] /= grid.integrate(dens * points**2 * np.exp(-e[0] * points**2) / approx * spherical)
    expons[1] /= grid.integrate(dens * points**4 * np.exp(-e[1] * points**2) / approx * spherical)
    assert_almost_equal(new_coeffs, c, decimal=6)
    assert_almost_equal(new_expons, expons, decimal=6)
    # test updating coeffs & expons
    new_coeffs, new_expons = kl._update_params(c, e, update_coeffs=True, update_expons=True)
    assert_almost_equal(new_coeffs, coeffs, decimal=6)
    assert_almost_equal(new_expons, expons, decimal=6)


def test_kl_scf_update_params_3d_molecular_dens_1s_1s_gaussian():
    r"""Test KL-SCF method in three dimensions for one s-type and one p-type Gaussian."""
    # actual density is a 1s Gaussian at origin
    axes = np.array([[0.1, 0.0, 0.0], [0.0, 0.1, 0.0], [0.0, 0.0, 0.1]])
    grid = CubicGrid(np.array([-2.0, -2.0, -2.0]), axes, (40, 40, 40))
    dens = np.exp(-np.sum(grid.points ** 2., axis=1))
    # model density is a normalized 1s Gassuain on each center
    coord = np.array([[0., 0., 0.], [0., 0., 1.]])
    c, e = np.array([1., 2.]), np.array([3., 4.])
    model = MolecularGaussianDensity(grid.points, coord, np.array([[1, 0], [1, 0]]), True)
    kl = KLDivergenceFPI(grid, dens, model)
    # compute expected updated coeffs
    dist1 = np.sum((grid.points - coord[0])**2, axis=1)
    dist2 = np.sum((grid.points - coord[1])**2, axis=1)
    approx = c[0] * (e[0] / np.pi)**1.5 * np.exp(-e[0] * dist1)
    approx += c[1] * (e[1] / np.pi)**1.5 * np.exp(-e[1] * dist2)
    expected_coeffs = c * (e / np.pi)**1.5
    expected_coeffs[0] *= grid.integrate(dens * np.exp(-e[0] * dist1) / approx)
    expected_coeffs[1] *= grid.integrate(dens * np.exp(-e[1] * dist2) / approx)
    # check updated coeffs
    coeffs, expons = kl._update_params(c, e, update_coeffs=True, update_expons=False)
    assert_almost_equal(expons, e, decimal=6)
    assert_almost_equal(coeffs, expected_coeffs, decimal=6)
    # compute expected updated expons
    expected_expons = 1.5 * (e / np.pi)**1.5
    expected_expons[0] *= grid.integrate(dens * np.exp(-e[0] * dist1) / approx)
    expected_expons[1] *= grid.integrate(dens * np.exp(-e[1] * dist2) / approx)
    denoms = (e / np.pi)**1.5
    denoms[0] *= grid.integrate(dens * dist1 * np.exp(-e[0] * dist1) / approx)
    denoms[1] *= grid.integrate(dens * dist2 * np.exp(-e[1] * dist2) / approx)
    expected_expons /= denoms
    # check updated expons
    coeffs, expons = kl._update_params(c, e, update_coeffs=False, update_expons=True)
    assert_almost_equal(coeffs, c, decimal=6)
    assert_almost_equal(expons, expected_expons, decimal=6)


def test_kl_scf_run_3d_molecular_dens_1s_1p_gaussian():
    r"""Test KL-SCF on 3D Gaussian example with 1 s-type and 1 p-type Gaussians."""
    # make cubic grid
    axes = np.array([[0.3, 0.0, 0.0], [0.0, 0.3, 0.0], [0.0, 0.0, 0.3]])
    grid = CubicGrid(np.array([-10.0, -10.0, -10.0]), axes, (65, 65, 65))
    # actual density is a 1s Gaussian at x=0.5 & 1p Gaussian at y=0.25
    c, e = np.array([0.53, 2.07]), np.array([0.67, 1.92])
    coord = np.array([[0.5, 0., 0.], [0., 0.25, 0.]])
    dist1 = np.sum((grid.points - coord[0])**2., axis=1)
    dist2 = np.sum((grid.points - coord[1])**2., axis=1)
    dens1 = c[0] * (e[0] / np.pi)**1.5 * np.exp(-e[0] * dist1)
    dens2 = c[1] * 2. * e[1]**2.5 * dist2 * np.exp(-e[1] * dist2) / (3. * np.pi**1.5)
    # model density is a normalized 1s Gaussian on each center
    model = MolecularGaussianDensity(grid.points, coord, np.array([[1, 0], [0, 1]]), True)
    # check model norm
    assert_almost_equal(grid.integrate(model.evaluate(c, e, False)), np.sum(c), decimal=8)
    # initial coeffs & expons for optimization
    cs0, es0 = np.array([0.25, 0.01]), np.array([0.5, 2.2])
    # optimize coeffs
    kl = KLDivergenceFPI(grid, dens1 + dens2, model, mask_value=0.)
    result = kl.run(cs0, e, True, False, 100, 1.e-8, 1.e-8, 1e-15)
    assert_equal(result["success"], True)
    assert_almost_equal(result["coeffs"], c, decimal=6)
    assert_almost_equal(result["exps"], e, decimal=6)
    assert_almost_equal(result["fun"][-1], 0., decimal=6)
    # optimize expons
    result = kl.run(c, es0, False, True, 100, 1.e-8, 1.e-8, 1.e-15)
    assert_equal(result["success"], True)
    assert_almost_equal(result["coeffs"], c, decimal=6)
    assert_almost_equal(result["exps"], e, decimal=6)
    assert_almost_equal(result["fun"][-1], 0., decimal=6)
    # optimize coeffs & expons
    result = kl.run(cs0, es0, True, True, 150, 1.e-7, 1.e-7, 1.e-15)
    assert_equal(result["success"], True)
    assert_almost_equal(result["coeffs"], c, decimal=6)
    assert_almost_equal(result["exps"], e, decimal=6)
    assert_almost_equal(result["fun"][-1], 0., decimal=6)


def test_kl_fit_unnormalized_dens_normalized_1s_gaussian():
    r"""Test ScipyFit against normalized gaussian model."""
    # density is normalized 1s orbital with exponent=1.0
    grid = UniformRadialGrid(150, 0.0, 15.0)
    # density is normalized 1s gaussian
    dens = 1.57 * np.exp(-0.51 * grid.points**2.)
    model = AtomicGaussianDensity(grid.points, num_s=1, num_p=0, normalize=True)
    measure = KLDivergence()
    kl = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # expected coeffs & expons
    expected_cs = np.array([1.57 / (0.51 / np.pi)**1.5])
    expected_es = np.array([0.51])
    # initial coeff=1.57 & expon=0.51
    result = kl.run(np.array([1.57]), np.array([0.51]), True, True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., 0.]), result["jacobian"], decimal=6)
    # initial coeff=0.1 & expon=0.1
    result = kl.run(np.array([0.1]), np.array([0.1]), True, True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., 0.]), result["jacobian"], decimal=6)
    # initial coeff=5.0 & expon=15.
    result = kl.run(np.array([5.0]), np.array([15.]), True, True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., 0.]), result["jacobian"], decimal=6)
    # initial coeff=0.8 & expon=0.51, opt coeffs
    result = kl.run(np.array([0.5]), np.array([0.51]), True, False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1.]), result["jacobian"], decimal=6)


def test_kl_fit_normalized_dens_unnormalized_1s_gaussian():
    r"""Test ScipyFit against un-normalized Gaussian model."""
    # density is normalized 1s gaussian
    grid = UniformRadialGrid(200, 0.0, 15.0)
    dens = 2.06 * (0.88 / np.pi)**1.5 * np.exp(-0.88 * grid.points**2.)
    # un-normalized 1s basis function
    model = AtomicGaussianDensity(grid.points, num_s=1, num_p=0, normalize=False)
    measure = KLDivergence()
    kl = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # expected coeffs & expons
    expected_cs = np.array([2.06]) * (0.88 / np.pi)**1.5
    expected_es = np.array([0.88])
    # initial coeff=2.6 & expon=0.001
    result = kl.run(np.array([2.6]), np.array([0.001]), opt_coeffs=True, opt_expons=True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=7)
    assert_almost_equal(expected_es, result["exps"], decimal=7)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=1. & expon=0.88, opt coeffs
    result = kl.run(np.array([1.0]), np.array([0.88]), opt_coeffs=True, opt_expons=False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1.]) / (0.88 / np.pi)**1.5, result["jacobian"], decimal=6)
    # initial coeff=10. & expon=0.88, opt coeffs
    result = kl.run(np.array([10.]), np.array([0.88]), opt_coeffs=True, opt_expons=False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1.]) / (0.88 / np.pi)**1.5, result["jacobian"], decimal=6)
    # initial coeff=expected_cs & expon=expected_es, opt expons
    result = kl.run(expected_cs, expected_es, opt_coeffs=False, opt_expons=True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=expected_cs & expon=5.0, opt expons
    result = kl.run(expected_cs, np.array([5.0]), opt_coeffs=False, opt_expons=True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_kl_fit_normalized_dens_normalized_1s_gaussian():
    r"""Test ScipyFiy againstnormalized Gaussian model."""
    # density is normalized 1s gaussian
    grid = UniformRadialGrid(150, 0.0, 15.0)
    dens = 2.06 * (0.88 / np.pi)**1.5 * np.exp(-0.88 * grid.points**2.)
    # normalized 1s basis function
    model = AtomicGaussianDensity(grid.points, num_s=1, num_p=0, normalize=True)
    measure = KLDivergence()
    kl = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # expected coeffs & expons
    expected_cs = np.array([2.06])
    expected_es = np.array([0.88])
    # initial coeff=2.6 & expon=0.001
    result = kl.run(np.array([2.6]), np.array([0.001]), opt_coeffs=True, opt_expons=True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., 0.]), result["jacobian"], decimal=6)
    # initial coeff=1. & expon=0.88, opt coeffs
    result = kl.run(np.array([1.0]), np.array([0.88]), opt_coeffs=True, opt_expons=False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1.]), result["jacobian"], decimal=6)
    # initial coeff=10. & expon=0.88, opt coeffs
    result = kl.run(np.array([10.]), np.array([0.88]), opt_coeffs=True, opt_expons=False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1.]), result["jacobian"], decimal=6)
    # initial coeff=expected_cs & expon=expected_es, opt expons
    result = kl.run(expected_cs, expected_es, opt_coeffs=False, opt_expons=True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([0.]), result["jacobian"], decimal=6)


def test_kl_fit_normalized_dens_unnormalized_2p_gaussian():
    r"""Test ScipyFiy against un-normalized p-type Gaussian."""
    # density is normalized 2p orbitals
    grid = UniformRadialGrid(150, 1e-15, 15.0)
    points = grid.points
    cs0 = np.array([0.76, 3.09])
    es0 = np.array([2.01, 0.83])
    dens = cs0[0] * es0[0]**2.5 * points**2 * np.exp(-es0[0] * points**2)
    dens += cs0[1] * es0[1]**2.5 * points**2 * np.exp(-es0[1] * points**2)
    dens *= 2. / (3. * np.pi**1.5)
    # un-normalized 2p functions
    model = AtomicGaussianDensity(points, num_s=0, num_p=2, normalize=False)
    measure = KLDivergence()
    kl = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # expected coeffs & expons
    expected_cs = cs0 * es0**2.5 * 2. / (3. * np.pi**1.5)
    expected_es = es0
    # initial coeff=[1.5, 0.1] & expon=[4., 0.001]
    result = kl.run(np.array([1.5, 0.1]), np.array([4.0, 0.001]), True, True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=cs0 & expon=es0, opt coeffs
    result = kl.run(cs0, es0, opt_coeffs=True, opt_expons=False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=[1.0, 1.0] & expon=es0, opt coeffs
    result = kl.run(np.array([1.0, 1.0]), es0, opt_coeffs=True, opt_expons=False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=expected_cs & expon=[4.5, 1.0], opt expons
    result = kl.run(expected_cs, np.array([4.5, 1.0]), opt_coeffs=False, opt_expons=True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_kl_fit_normalized_dens_normalized_2p_gaussian():
    r"""Test ScipyFiy against normalized p-type Gaussian model."""
    # density is normalized 2p orbitals
    grid = UniformRadialGrid(150, 1e-15, 15.0)

    points = grid.points
    cs0 = np.array([0.76, 3.09])
    es0 = np.array([2.01, 0.83])
    dens = cs0[0] * es0[0]**2.5 * points**2 * np.exp(-es0[0] * points**2)
    dens += cs0[1] * es0[1]**2.5 * points**2 * np.exp(-es0[1] * points**2)
    dens *= 2. / (3. * np.pi ** 1.5)
    # normalized 2p functions
    model = AtomicGaussianDensity(points, num_s=0, num_p=2, normalize=True)
    measure = KLDivergence()
    kl = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # initial coeff=cs0 & expon=es0, opt coeffs
    result = kl.run(cs0, es0, True, False)
    assert_almost_equal(cs0, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., -1.]), result["jacobian"], decimal=6)
    # initial coeff=[2.5, 0.5] & expon=[2.0, 1.9]
    result = kl.run(np.array([2.5, 0.5]), np.array([2.0, 1.9]), True, True)
    assert_almost_equal(np.sort(cs0), np.sort(result["coeffs"]), decimal=6)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., -1., 0., 0.]), result["jacobian"], decimal=6)
    # initial coeff=[1.0, 1.0] & expon=es0, opt coeffs
    result = kl.run(np.array([1.0, 1.0]), es0, True, False)
    assert_almost_equal(cs0, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., -1.]), result["jacobian"], decimal=6)
    # initial coeff=cs0 & expon=es0, opt coeffs
    result = kl.run(cs0, es0, True, False)
    assert_almost_equal(cs0, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., -1.]), result["jacobian"], decimal=6)


def test_kl_fit_normalized_dens_normalized_1s2p_gaussian():
    r"""Test ScipyFit with KL against normalized one s-type and 2 p-type Gaussians."""
    # density is normalized 1s + 2p gaussians
    grid = UniformRadialGrid(150, 0.0, 15.0)
    points = grid.points
    cs0 = np.array([1.52, 0.76, 3.09])
    es0 = np.array([0.50, 2.01, 0.83])
    dens = cs0[0] * (es0[0] / np.pi)**1.5 * np.exp(-es0[0] * points**2.)
    dens += cs0[1] * 2 * es0[1]**2.5 * points**2 * np.exp(-es0[1] * points**2) / (3. * np.pi**1.5)
    dens += cs0[2] * 2 * es0[2]**2.5 * points**2 * np.exp(-es0[2] * points**2) / (3. * np.pi**1.5)
    # un-normalized 1s + 2p functions
    model = AtomicGaussianDensity(points, num_s=1, num_p=2, normalize=True)
    measure = KLDivergence()
    kl = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # initial coeff=1. & expon=1.
    result = kl.run(np.array([1., 1., 1.]), np.array([1., 1., 1.]), True, True)
    assert_almost_equal(np.sort(cs0), np.sort(result["coeffs"]), decimal=5)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., -1., -1., 0., 0., 0.]), result["jacobian"], decimal=6)
    # initial coeff=[0.1, 0.6, 7.] & expon=[1., 0.9, 1.0]
    result = kl.run(np.array([0.1, 0.6, 7.]), np.array([1., 0.9, 1.0]), True, True)
    assert_almost_equal(np.sort(cs0), np.sort(result["coeffs"]), decimal=5)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., -1., -1., 0., 0., 0.]), result["jacobian"], decimal=6)
    # initial coeff=[1., 5., 0.] & expon=es0, opt coeffs
    result = kl.run(np.array([1., 5., 0.]), es0, True, False)
    assert_almost_equal(np.sort(cs0), np.sort(result["coeffs"]), decimal=5)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., -1., -1.]), result["jacobian"], decimal=6)
    # initial coeff=cs0 & expon=es0, opt expons
    result = kl.run(cs0, es0, False, True)
    assert_almost_equal(np.sort(cs0), np.sort(result["coeffs"]), decimal=5)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([0., 0., 0.]), result["jacobian"], decimal=6)


def test_kl_fit_unnormalized_1d_molecular_dens_unnormalized_1s_1s_gaussian():
    r"""Test ScipyFit of 3D un-normalized Gaussian model."""
    # density is normalized 1s + 1s gaussians
    grid = UniformRadialGrid(150, 0.0, 15.0)
    points = grid.points
    cs0 = np.array([1.52, 2.67])
    es0 = np.array([0.31, 0.41])
    coords = np.array([[0.], [1.]])
    # compute density on each center
    dens1 = cs0[0] * np.exp(-es0[0] * (points - coords[0])**2.)
    dens2 = cs0[1] * np.exp(-es0[1] * (points - coords[1])**2.)
    # un-normalized 1s + 1s functions
    model = MolecularGaussianDensity(points, coords, np.array([[1, 0], [1, 0]]), False)
    # fit total density
    measure = KLDivergence()
    kl = ScipyFit(grid, dens1 + dens2, model, measure=measure, method="slsqp", spherical=True)
    # initial coeff=1. & expon=1.
    result = kl.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(cs0, result["coeffs"], decimal=4)
    assert_almost_equal(es0, result["exps"], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=1. & expon=1.
    result = kl.run(np.array([5.45, 0.001]), es0, True, False, disp=True)
    assert_almost_equal(cs0, result["coeffs"], decimal=5)
    assert_almost_equal(es0, result["exps"], decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=1. & expon=1.
    result = kl.run(cs0, np.array([5.45, 0.001]), False, True)
    assert_almost_equal(cs0, result["coeffs"], decimal=5)
    assert_almost_equal(es0, result["exps"], decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    # fit 1s density on center 1
    measure = KLDivergence()
    kl = ScipyFit(grid, dens1, model, measure=measure, method="slsqp", spherical=True)
    # initial coeff=1. & expon=1.
    result = kl.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(np.array([result["coeffs"][0], 0.]), result["coeffs"], decimal=4)
    assert_almost_equal(es0[0], result["exps"][0], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=1. & expon=1.
    result = kl.run(np.array([6.79, 8.51]), es0, True, False)
    assert_almost_equal(np.array([result["coeffs"][0], 0.]), result["coeffs"], decimal=4)
    assert_almost_equal(es0, result["exps"], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=1. & expon=1.
    result = kl.run(np.array([1.52, 0.0]), np.array([3.0, 4.98]), False, True)
    assert_almost_equal(np.array([result["coeffs"][0], 0.]), result["coeffs"], decimal=4)
    assert_almost_equal(es0[0], result["exps"][0], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_kl_fit_unnormalized_1d_molecular_dens_unnormalized_1s_1p_gaussian():
    r"""Test ScipyFit against un-normalized 1S and 1P-type Gaussians."""
    # density is normalized 1s + 1s gaussians
    grid = UniformRadialGrid(150, 0.0, 15.0)
    points = grid.points
    cs0 = np.array([1.52, 2.67])
    es0 = np.array([0.31, 0.41])
    coords = np.array([[0.], [1.]])
    # compute density of each center
    dens_s = cs0[0] * (es0[0] / np.pi)**1.5 * np.exp(-es0[0] * (points - coords[0])**2.)
    dens_p = cs0[1] * (points - coords[1])**2 * np.exp(-es0[1] * (points - coords[1])**2.)
    dens_p *= (2. * es0[1]**2.5 / (3. * np.pi**1.5))
    # un-normalized 1s + 1p functions
    model = MolecularGaussianDensity(points, coords, np.array([[1, 0], [0, 1]]), True)
    # fit total density
    measure = KLDivergence()
    kl = ScipyFit(grid, dens_s + dens_p, model, measure=measure, method="slsqp", spherical=True)
    # opt. coeffs & expons
    result = kl.run(np.ones(2), np.array([0.1, 0.2]), True, True, tol=1e-20, disp=True)
    assert_almost_equal(cs0, result["coeffs"], decimal=4)
    assert_almost_equal(es0, result["exps"], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. coeffs, initial expon=es0
    result = kl.run(np.array([5.91, 7.01]), es0, True, False)
    assert_almost_equal(cs0, result["coeffs"], decimal=4)
    assert_almost_equal(es0, result["exps"], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. expons, initial coeff=cs0
    result = kl.run(cs0, np.array([5.91, 7.01]), False, True)
    assert_almost_equal(cs0, result["coeffs"], decimal=4)
    assert_almost_equal(es0, result["exps"], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)
    # fit 1s density on center 1
    measure = KLDivergence()
    kl = ScipyFit(grid, dens_s, model, measure=measure, method="slsqp", spherical=True)
    # opt. coeffs & expons
    result = kl.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(np.array([cs0[0], 0.]), result["coeffs"], decimal=4)
    assert_almost_equal(es0[0], result["exps"][0], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)
    # # fit 1p density on center 2
    measure = KLDivergence(mask_value=1e-12)
    kl = ScipyFit(grid, dens_p, model, measure=measure, method="slsqp", spherical=True)
    # opt. expons
    result = kl.run(np.array([0., cs0[1]]), np.ones(2), False, True)
    assert_almost_equal(np.array([0., cs0[1]]), result["coeffs"], decimal=4)
    assert_almost_equal(es0[1], result["exps"][1], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_ls_fit_normalized_dens_normalized_1s_gaussian():
    r"""Test ScipyFit of least-squares against normalized Gaussian model."""
    # density is normalized 1s orbital with exponent=1.0
    grid = UniformRadialGrid(200, 0.0, 15.0)
    # actual density is a normalized 1s gaussian
    cs0, es0 = np.array([1.57]), np.array([0.51])
    dens = 1.57 * (0.51 / np.pi)**1.5 * np.exp(-0.51 * grid.points**2.)
    # model density is a normalized 1s Gaussian
    model = AtomicGaussianDensity(grid.points, num_s=1, num_p=0, normalize=True)
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # opt. coeffs & expons
    result = ls.run(np.array([0.1]), np.array([3.5]), True, True)
    assert_almost_equal(cs0, result["coeffs"], decimal=8)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. coeffs
    result = ls.run(np.array([10.]), np.array([0.51]), True, False)
    assert_almost_equal(cs0, result["coeffs"], decimal=8)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # model density is two normalized 1s Gaussian
    model = AtomicGaussianDensity(grid.points, num_s=2, num_p=0, normalize=True)
    ls = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # opt. coeffs & expons
    result = ls.run(np.array([0.1, 3.0]), np.array([5.1, 7.8]), True, True)
    assert_almost_equal(np.array([1.57, 0.]), result["coeffs"], decimal=6)
    assert_almost_equal(np.array([0.51, 0.]), result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_ls_fit_normalized_dens_normalized_2s_gaussian():
    r"""Test ScipyFit least-squares against two normalized s-type Gaussians."""
    # density is normalized 1s orbital with exponent=1.0
    grid = UniformRadialGrid(300, 0.0, 15.0)
    # actual density is a normalized 1s gaussian
    cs0 = np.array([1.57, 0.12])
    es0 = np.array([0.45, 1.29])
    # evaluate normalized 8s density
    dens = cs0[0] * (es0[0] / np.pi)**1.5 * np.exp(-es0[0] * grid.points**2.)
    dens += cs0[1] * (es0[1] / np.pi)**1.5 * np.exp(-es0[1] * grid.points**2.)
    # check norm of density
    assert_almost_equal(
        grid.integrate(dens * 4.0 * np.pi * grid.points**2.0), np.sum(cs0), decimal=6
    )
    # model density is a normalized 2s Gaussian
    model = AtomicGaussianDensity(grid.points, num_s=2, num_p=0, normalize=True)
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    initial_cs = np.array([0.57, 0.98])
    initial_es = np.array([1.67, 0.39])
    # opt. coeffs & expons
    result = ls.run(initial_cs, initial_es, True, True)
    assert_almost_equal(np.sort(cs0), np.sort(result["coeffs"]), decimal=6)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. coeffs
    result = ls.run(initial_cs, es0, True, False)
    assert_almost_equal(cs0, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_ls_fit_normalized_dens_normalized_5s_gaussian():
    r"""Test ScipyFiy least-squares against 5 s-type, normalized Gaussians."""
    # density is normalized 1s orbital with exponent=1.0
    grid = UniformRadialGrid(300, 0.0, 15.0)
    # actual density is a normalized 5s gaussian
    cs0 = np.array([1.57, 0.12, 3.67, 0.97, 5.05])
    es0 = np.array([0.45, 1.29, 1.25, 20.1, 10.5])
    # evaluate normalized 5s density
    dens = cs0[0] * (es0[0] / np.pi)**1.5 * np.exp(-es0[0] * grid.points**2)
    dens += cs0[1] * (es0[1] / np.pi)**1.5 * np.exp(-es0[1] * grid.points**2)
    dens += cs0[2] * (es0[2] / np.pi)**1.5 * np.exp(-es0[2] * grid.points**2)
    dens += cs0[3] * (es0[3] / np.pi)**1.5 * np.exp(-es0[3] * grid.points**2)
    dens += cs0[4] * (es0[4] / np.pi)**1.5 * np.exp(-es0[4] * grid.points**2)
    # check norm of density
    assert_almost_equal(
        grid.integrate(dens * 4.0 * np.pi * grid.points**2.0), np.sum(cs0), decimal=6
    )
    # model density is a normalized 1s Gaussian
    model = AtomicGaussianDensity(grid.points, num_s=5, num_p=0, normalize=True)
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    initial_cs = np.array([1., 0., 3., 0., 5.])
    initial_es = np.array([0.1, 1.2, 1., 20., 10.])
    # opt. coeffs
    result = ls.run(initial_cs, es0, True, False, tol=1e-15)
    assert_almost_equal(np.sort(cs0), np.sort(result["coeffs"]), decimal=6)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. expons, by re-arranging the exponents and coefficients.
    initial_cs = np.array([0.12, 0.97,  3.67, 1.57, 5.05])
    initial_es = np.array([1.2, 19., 1., 0.1, 10.])
    result = ls.run(initial_cs, initial_es, False, True, tol=1e-20, with_constraint=False)
    assert_almost_equal(np.sort(initial_cs), np.sort(result["coeffs"]), decimal=6)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=1)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_ls_fit_unnormalized_1d_molecular_dens_unnormalized_2s_gaussian():
    r"""Test ScipyFit least-squares against un-normalized Gaussian model."""
    # density is normalized 1s + 1s gaussians
    grid = UniformRadialGrid(200, 0.0, 15.0)
    cs0, es0 = np.array([1.52, 2.67, ]), np.array([0.31, 0.41])
    coords = np.array([[0.], [1.]])
    # compute density on each center
    dens1 = cs0[0] * np.exp(-es0[0] * (grid.points - coords[0])**2.)
    dens2 = cs0[1] * np.exp(-es0[1] * (grid.points - coords[1])**2.)
    # un-normalized 1s + 1s basis functions
    model = MolecularGaussianDensity(grid.points, coords, np.array([[1, 0], [1, 0]]), False)
    # fit total density
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens1 + dens2, model, measure=measure, method="slsqp", spherical=True)
    # initial coeff=1. & expon=1.
    result = ls.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(cs0, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # fit 1s density on center 1
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens1, model, measure=measure, method="slsqp", spherical=True)
    # initial coeff=1. & expon=1.
    result = ls.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(np.array([1.52, 0.0]), result["coeffs"], decimal=6)
    assert_almost_equal(es0[0], result["exps"][0], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # fit 1s density on center 2
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens2, model, measure=measure, method="slsqp", spherical=True)
    # initial coeff=1. & expon=1.
    result = ls.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(np.array([0.0, 2.67]), result["coeffs"], decimal=6)
    assert_almost_equal(es0[1], result["exps"][1], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_ls_fit_normalized_1d_molecular_dens_unnormalized_1s_1p_gaussian():
    r"""Test ScipyFit least-squares of 3D 1S, 1P-type Gaussian model."""
    # density is normalized 1s + 1s gaussians
    grid = UniformRadialGrid(200, 0.0, 15.0)
    cs0, es0 = np.array([1.52, 2.67]), np.array([0.31, 0.41])
    coords = np.array([[0.0], [1.0]])
    # compute density of each center
    dens_s = cs0[0] * (es0[0] / np.pi)**1.5 * np.exp(-es0[0] * (grid.points - coords[0])**2.)
    dens_p = cs0[1] * (grid.points - coords[1])**2 * np.exp(-es0[1] * (grid.points - coords[1])**2)
    dens_p *= (2. * es0[1]**2.5 / (3. * np.pi**1.5))
    # normalized 1s + 1p basis functions
    model = MolecularGaussianDensity(grid.points, coords, np.array([[1, 0], [0, 1]]), False)
    expected_cs = cs0 * np.array([(es0[0] / np.pi)**1.5, 2. * es0[1]**2.5 / (3. * np.pi**1.5)])
    # fit total density
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens_s + dens_p, model, measure=measure, method="slsqp", spherical=True)
    # opt. coeffs & expons
    result = ls.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. coeffs
    result = ls.run(np.array([0.001, 3.671]), es0, True, False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. expons
    result = ls.run(expected_cs, np.array([2.5, 1.9]), False, True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # fit 1s density on center 1
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens_s, model, measure=measure, method="slsqp", spherical=True)
    # opt. coeffs & expons
    result = ls.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(np.array([expected_cs[0], 0.]), result["coeffs"], decimal=6)
    assert_almost_equal(es0[0], result["exps"][0], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. coeffs
    result = ls.run(np.array([0.05, 5.01]), es0, True, False)
    assert_almost_equal(np.array([expected_cs[0], 0.]), result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # fit 1s density on center 1
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens_p, model, measure=measure, method="slsqp", spherical=True)
    result = ls.run(np.array([0.05, 5.01]), es0, True, False, with_constraint=False)
    assert_almost_equal([0., expected_cs[1]], result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)



# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
r"""Obtain UGBS exponents from ygbs file."""
import inspect
import os

__all__ = ["get_ugbs_exponents"]


def get_ugbs_exponents(element):
    r"""
    Get the UGBS exponents of a element.

    Parameters
    ----------
    element : str
        The element whose UGBS exponents are required.

    Returns
    -------
    dict
        Dictionary with keys "S" or "P" type orbitals and the items are the exponents for that
        shell.

    """
    assert isinstance(element, str)
    path_to_function = os.path.abspath(inspect.getfile(get_ugbs_exponents))
    path_to_function = path_to_function[:-13]  # Remove parse_ugbs.py <- 13 characters
    file_path = path_to_function + r"data/ygbs"
    output = {"S" : [], "P" : [], "D" : [], "F" : []}
    with open(file_path, "r", encoding="utf8") as f:
        for line in f:
            split_words = line.strip().split()
            if len(split_words) > 0 and split_words[0].lower() == element.lower():
                next_line = f.readline().strip().split()
                output[split_words[1]].append(float(next_line[0]))
    return output



# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
r"""Fitting Algorithms."""

import warnings
from timeit import default_timer as timer

import numpy as np
from scipy.optimize import minimize, NonlinearConstraint

from bfit.measure import KLDivergence, Measure, SquaredDifference

__all__ = ["KLDivergenceFPI", "ScipyFit"]


class _BaseFit:
    r"""Base Fitting Class."""

    def __init__(self, grid, density, model, measure, integral_dens=None, spherical=False,
                 mask_value=1e-18):
        r"""
        Construct the base fitting class.

        Parameters
        ----------
        grid : (_BaseRadialGrid, CubicGrid)
            The grid class that contains the grid points and a integrate function.
             Located in `grid.py`
        density : ndarray(N,)
            The true density evaluated on :math:`N` grid points.
        model : (AtomicGaussianDensity, MolecularGaussianDensity)
            The Gaussian basis model density. See `model.py`.
        measure : (SquaredDifference, KLDivergence)
            The deviation measure between true density and model density. See `measure.py`.
        integral_dens : float, optional
            If this is provided, then the model is constrained to integrate to this value.
            If not, then the model is constrained to the numerical integration of the
            density. Useful when one knows the actual integration value of the density.
        spherical : bool, optional
            Whether to perform spherical integration by adding :math:`4 \pi r^2` term
            to the integrand. Only used when grid is one-dimensional and positive (radial grid).
        mask_value : float, optional
            Mask value used for calculating the Kullback-Leibler divergence. This value sets
            :math:`\log(f(x) / g(x)) = 0` when `g(x)` is less than the mask value.

        """
        if np.any(density < 0.):
            raise ValueError("Density should be positive.")
        self._grid = grid
        if not hasattr(self.grid, "points"):
            raise AttributeError(
                f"Grid class {type(self.grid)} should contain attribute 'points'."
            )
        if not (hasattr(self.grid, 'integrate') and callable(getattr(self.grid, 'integrate'))):
            raise AttributeError(
                f"Grid class {type(self.grid)} should contain method called 'integrate'."
            )
        if self.grid.points.ndim != 1 and spherical:
            raise ValueError(
                f"Spherical is true only when grid points {self.grid.points.ndim} are "
                f"one-dimensional."
            )
        self._density = density
        self._model = model
        self._measure = measure
        self._spherical = spherical
        # compute norm of density
        if integral_dens is None:
            self._integral_dens = self.integrate(density)
        else:
            self._integral_dens = integral_dens

        # Used to calculate the error measures in model only.
        self.kl_error = KLDivergence(mask_value=mask_value)
        self.ls_error = SquaredDifference()

    @property
    def grid(self):
        r"""Return grid object containing points and integration method."""
        return self._grid

    @property
    def density(self):
        r"""Return the true density evaluated on the grid points."""
        return self._density

    @property
    def model(self):
        r"""Return the Gaussian basis model density."""
        return self._model

    @property
    def measure(self):
        r"""Return the deviation measure between true density and model density."""
        return self._measure

    @property
    def integral_dens(self):
        r"""Return integration value of the density."""
        return self._integral_dens

    @property
    def spherical(self):
        """Whether to perform spherical integration."""
        return self._spherical

    def integrate(self, integrand):
        r"""
        Integrate the integrand.

        Parameters
        ----------
        integrand : ndarray(N,)
            The integrand :math:`f(x)` being integrated defined on :math:`N` points.
            If `spherical` attribute is true, then the radial component is
            integrated in spherical coordinates.

        Returns
        -------
        float :
            Returns the integration :math:`\int f(x) dx` or if spherical
            is true then returns :math:`\int_0^\infty f(r) 4 \pi r^2 dr`.

        """
        if self.spherical:
            return self.grid.integrate(integrand * 4.0 * np.pi * self.grid.points**2.0)
        return self.grid.integrate(integrand)

    def goodness_of_fit(self, coeffs, expons):
        r"""
        Compute various measures over the grid to determine the accuracy of the fitted model.

        In particular, it computes the integral of the model, the :math:`L_1` distance,
        the :math:`L_\infty` distance and attribute `measure` distance between true and model
        functions.

        Parameters
        ----------
        coeffs : ndarray
            The coefficients of Gaussian basis functions.
        expons : ndarray
            The exponents of Gaussian basis functions.

        Returns
        -------
        integral : float
            Integral of approximate model density, i.e. norm of approximate model density.
        l_1 : float
            Integral of absolute difference between density and approximate model density.
            This is defined to be :math:`L_(f, g) = \int |f(x) - g(x)| dx`.
        l_infinity : float
            The maximum absolute difference between density and approximate model density.
            This is defined to be :math:`L_\infty(f, g) = \max |f(x) - g(x)|`.
        least_squares : float
            Square of the :math:`L_2` norm between density and approximate model density.
            This is defined to be :math:`L_2^2(f, g) = \int (f(x) - g(x))^2 dx`.
        kullback_leibler : float
            Kullback-Leibler divergence between density and approximate model density.
            This is defined to be :math:`KL(f, g) = \int f(x) \log\bigg(\frac{f(x)}{g(x)}\bigg)dx`.
            If :math:`g` is negative, then this returns infinity.

        """
        # evaluate approximate model density
        approx = self.model.evaluate(coeffs, expons)
        diff = np.abs(self.density - approx)
        return [
            self.integrate(approx),
            self.integrate(diff),
            np.max(diff),
            self.integrate(self.ls_error.evaluate(self.density, approx, deriv=False)),
            self.integrate(self.kl_error.evaluate(self.density, approx, deriv=False))
        ]


class KLDivergenceFPI(_BaseFit):
    r"""
    Kullback-Leibler Divergence FPI method for fitting to a density.

    This class optimizes the following objective function using a fixed point iteration method,

    .. math::
        \min_{\{c_i\}, \{\alpha\}} \int f(x) \log \bigg(\frac{f(x)}{\sum c_i b_i(x, \alpha_i)}
         \bigg)dx + \lambda(N - \sum c_i)

    where :math:`f` is the true density to be fitted,
    :math:`\lambda` is the Lagrange multiplier,
    :math:`c_i` is the coefficient of the ith basis function
    that all sum to the integral of density function :math:`N= \int f(x)dx`, and
    :math:`\alpha_i` is the exponent of the basis function :math:`b_i`.

    """

    def __init__(self, grid, density, model, mask_value=0., integral_dens=None, spherical=False):
        r"""
        Construct the KLDivergenceFPI class.

        Parameters
        ----------
        grid : (_BaseRadialGrid, CubicGrid)
            Grid class that contains the grid points and integration methods on them.
        density : ndarray
            The true density evaluated on the grid points.
        model : (AtomicGaussianDensity, MolecularGaussianDensity)
            The Gaussian basis model density. Located in `model.py`.
        mask_value : float, optional
            The elements less than or equal to this number are masked in a division.
        integral_dens : float, optional
            If this is provided, then the model is constrained to integrate to this value.
            If not, then the model is constrained to the numerical integration of the
            density. Useful when one knows the actual integration value of the density.
        spherical : bool
            Whether to perform spherical integration by adding :math:`4 \pi r^2` term
            to the integrand. Only used when grid is one-dimensional and positive (radial grid).

        """
        # initialize KL deviation measure
        measure = KLDivergence(mask_value=mask_value)
        super().__init__(grid, density, model, measure, integral_dens, spherical, mask_value)
        # compute lagrange multiplier
        self._lm = self.integrate(self.density) / self.integral_dens
        if self._lm == 0. or np.isnan(self._lm):
            raise RuntimeError(f"Lagrange multiplier cannot be {self._lm}.")

    @property
    def lagrange_multiplier(self):
        """Lagrange multiplier of Kullback-Leibler optimization problem."""
        return self._lm

    def _update_params(self, coeffs, expons, update_coeffs=True, update_expons=False):
        r"""
        Update coefficients & exponents of the Gaussian density model.

        Parameters
        ----------
        coeffs : ndarray
            The initial coefficients of Gaussian basis functions.
        expons : ndarray
            The initial exponents of Gaussian basis functions.
        update_coeffs : bool, optional
            Whether to optimize coefficients of Gaussian basis functions.
            Default is true.
        update_expons : bool, optional
            Whether to optimize exponents of Gaussian basis functions.
            Default is true.

        Returns
        -------
        coeffs : ndarray
            The updated coefficients of Gaussian basis functions. Only returned if
            `update_coeffs=True`.
        expons : ndarray
            The updated exponents of Gaussian basis functions. Only returned if
            `update_expones=True`.

        """
        if not update_coeffs and not update_expons:
            raise ValueError("At least one of args update_coeff or update_expons should be True.")
        # compute model density & its derivative
        m, dm = self.model.evaluate(coeffs, expons, deriv=True)
        # compute KL divergence & its derivative
        _, dk = self.measure.evaluate(self.density, m, deriv=True)
        # compute averages needed to update parameters
        avrg1, avrg2 = np.zeros(self.model.nbasis), np.zeros(self.model.nbasis)
        for index in range(self.model.nbasis):
            integrand = -dk * dm[:, index]
            avrg1[index] = self.integrate(integrand)
            if update_expons:
                if self.model.natoms == 1:
                    # case of AtomicGaussianDensity or MolecularGaussianDensity model with 1 atom
                    radii = np.ravel(self.model.radii)
                else:
                    # case of MolecularGaussianDensity model with more than 1 atom
                    center_index = self.model.assign_basis_to_center(index)
                    radii = self.model.radii[center_index]
                avrg2[index] = self.integrate(integrand * radii**2)

        # compute updated coeffs & expons
        if update_coeffs:
            coeffs = coeffs * avrg1 / self._lm
        if update_expons:
            expons = self.model.prefactor * avrg1 / avrg2
        return coeffs, expons

    def run(self, c0, e0, opt_coeffs=True, opt_expons=True, maxiter=500, c_threshold=1.e-6,
            e_threshold=1.e-6, d_threshold=1.e-6, disp=False):
        r"""
        Optimize the coefficients & exponents of Gaussian basis functions via fixed-point.

        Parameters
        ----------
        c0 : ndarray
            The initial coefficients of Gaussian basis functions.
        e0 : ndarray
            The initial exponents of Gaussian basis functions.
        opt_coeffs : bool, optional
            Whether to optimize coefficients of Gaussian basis functions.
            Default is true.
        opt_expons : bool, optional
            Whether to optimize exponents of Gaussian basis functions.
            Default is true.
        maxiter : int, optional
            Maximum number of iterations.
        c_threshold : float
            The termination threshold for absolute change in coefficients. Default is 1e-6.
        e_threshold : float
            The termination threshold for absolute change in exponents. Default is 1e-6.
        d_threshold : float
            The termination threshold for absolute change in divergence value. Default is 1e-6.
        disp : bool
            If true, then at each iteration various error measures will be printed.

        Returns
        -------
        dict :
            The optimization results presented as a dictionary with keys:

            "coeffs" : ndarray
                The optimized coefficients of the Gaussian model.
            "exps" : ndarray
                The optimized exponents of the Gaussian model.
            "success": bool
                Whether the optimization exited successfully.
            "fun" : ndarray
                Values of the KL divergence (objective function) at each iteration.
            "performance" : ndarray
                Values of various performance measures of modeled density at each iteration,
                as computed by `goodness_of_fit()` method.
            "time" : float
                The time in seconds it took to complete the algorithm.

        """
        # check the shape of initial coeffs and expons
        if not isinstance(c0, np.ndarray) or not isinstance(e0, np.ndarray):
            raise TypeError("Initial coefficients or exponents should be numpy arrays.")
        if c0.shape != (self.model.nbasis,):
            raise ValueError(f"Argument init_coeffs shape != ({self.model.nbasis},)")
        if e0.shape != (self.model.nbasis,):
            raise ValueError(f"Argument init_expons shape != ({self.model.nbasis},)")

        new_cs, new_es = c0, e0

        diff_divergence = np.inf
        max_diff_coeffs = np.inf
        max_diff_expons = np.inf

        fun, performance = [], []
        niter = 0
        start = timer()

        if disp:
            # Template for the header.
            print("-" * (10 + 12 + 15 + 15 + 15 + 15 + 15 + 15 + 15 + 12))
            # Format is {identifier:width}, ^ means center it, | means put a bar in it
            template_header = (
                "|{0:^10}|{1:^12}|{2:^15}|{3:^15}|{4:^15}|{5:^16}|{6:^15}|{7:^15}|{8:^15}|"
            )
            # Print the headers
            print(template_header.format(
                "Iteration", "Integration", "L1", "L Infinity", "Least-squares",
                "Kullback-Leibler", "Change Coeffs", "Change Exps", "Change Objective")
            )
            print("-" * (10 + 12 + 15 + 15 + 15 + 15 + 15 + 15 + 15 + 12))
            # Template for float iteration
            template_iters = (
                "|{0:^10d}|{1:^12f}|{2:^15e}|{3:^15f}|{4:^15e}|{5:^16e}|{6:^15e}|{7:^15e}|{8:^15e}|"
            )
        while ((max_diff_expons > e_threshold or max_diff_coeffs > c_threshold) and
               diff_divergence > d_threshold) and maxiter > niter:
            # update old coeffs & expons
            old_cs, old_es = new_cs, new_es
            # update coeffs and/or exponents
            if opt_coeffs and opt_expons:
                new_cs, new_es = self._update_params(new_cs, new_es, True, True)
            elif opt_coeffs:
                new_cs, new_es = self._update_params(new_cs, new_es, True, False)
            elif opt_expons:
                new_cs, new_es = self._update_params(new_cs, new_es, False, True)
            else:
                raise ValueError("Both opt_coeffs & opt_expons are False! Nothing to optimize!")
            # compute max change in cs & expons
            max_diff_coeffs = np.max(np.abs(new_cs - old_cs))
            max_diff_expons = np.max(np.abs(new_es - old_es))
            # compute errors & update niter
            performance.append(self.goodness_of_fit(new_cs, new_es))
            fun.append(performance[-1][-1])
            niter += 1

            # compute absolute change in divergence
            if niter != 1:
                diff_divergence = np.abs(performance[niter - 1][-1] - performance[niter - 2][-1])

            if disp:
                print(template_iters.format(
                     niter, *performance[-1], max_diff_coeffs, max_diff_expons, diff_divergence)
                )

        end = timer()
        time = end - start

        # check whether convergence is reached.
        if (maxiter == niter and diff_divergence > d_threshold) or np.isnan(diff_divergence):
            success = False
        else:
            success = True

        if disp:
            print("-" * (10 + 12 + 15 + 15 + 15 + 15 + 15 + 15 + 15 + 12))
            print(f"Successful?: {success}")
            print(f"Time: {time:.2f} seconds")
            print(f"Coefficients {new_cs}")
            print(f"Exponents {new_es}")

        results = {"coeffs": new_cs,
                   "exps" : new_es,
                   "fun": np.array(fun),
                   "success": success,
                   "performance": np.array(performance),
                   "time": time}

        return results


class ScipyFit(_BaseFit):
    r"""
    Optimize least-squares or Kullback-Leibler of Gaussian functions using `Scipy.optimize`.

    Least-squares objective function w.r.t. Gaussian basis-functions is defined as

    .. math::
        \min_{\{c_i\}, \{\alpha\}} \int \bigg(f(x) - \sum c_i b_k(x, \alpha_i) \bigg)^2 dx

    The Kullback-Leibler divergence function is defined as

    .. math::
        \min_{\{c_i\}, \{\alpha\}, \sum c_i = N} \int f(x) \log \bigg(\frac{f(x)}{\sum c_i
        b_i(x, \alpha_i)} \bigg)dx,

    where :math:`f` is the density to be fitted to,
    :math:`c_i, \alpha_i` are the coefficeints and exponents of the Gaussian
    basis functions, and :math:`N = \int f(x)dx` is the integral of the density function.

    Notes
    -----
    - Note that the Kullback-Leibler between two functions :math:`f` and :math:`g` is positive
      if and only if the integrals of :math:`f` and :math:`g` are identical.  The `with_constraint`
      attribute must be True for optimizing Kullback-Leibler.

    """

    def __init__(self, grid, density, model, measure=KLDivergence, method="SLSQP", weights=None,
                 integral_dens=None, spherical=False, mask_value=1e-18):
        r"""
        Construct the ScipyFit object.

        Parameters
        ----------
        grid : (_BaseRadialGrid, CubicGrid)
            The grid class.
        density : ndarray(N,)
            The true density evaluated on the grid points.
        model : (AtomicGaussianDensity, MolecularGaussianDensity)
            The Gaussian basis model density.
        measure : bfit.measure.Measure
            The deviation measure between true density and model density.
            See bfit.measure.py for examples of measures to use.
        method : str, optional
            The method used for optimizing parameters. Default is "slsqp".
            See "scipy.optimize.minimize" for options.
        weights : ndarray, optional
            The weights of objective function at each point. If `None`, 1.0 is used.
        integral_dens : float, optional
            If this is provided, then the model is constrained to integrate to this value.
            If not, then the model is constrained to the numerical integration of the
            density. Useful when one knows the actual integration value of the density.
        spherical : bool, optional
            Whether to perform spherical integration by adding :math:`4 \pi r^2` term
            to the integrand. Only used when grid is one-dimensional and positive (radial grid).
        mask_value : float, optional
            Mask value used for calculating the Kullback-Leibler divergence. This value sets
            :math:`\log(f(x) / g(x)) = 0` when `g(x)` is less than the mask value.

        """
        if np.any(abs(grid.points - model.points) > 1.e-12):
            raise ValueError("The grid.points & model.points are not the same!")
        if len(grid.points) != len(density):
            raise ValueError(f"Argument density should have ({len(grid.points)},) shape.")
        if method.lower() not in ["slsqp", "trust-constr"]:
            raise ValueError(f"Argument method={method} is not recognized!")
        if not isinstance(measure, Measure):
            raise TypeError(f"Measure {type(measure)} needs to be a children of the class Measure.")

        self.method = method.lower()
        # Assign the weights.
        if weights is None:
            weights = np.ones(len(density))
        self.weights = weights
        super().__init__(grid, density, model, measure, integral_dens, spherical, mask_value)

    def run(self, c0, e0, opt_coeffs=True, opt_expons=True, maxiter=1000, tol=1.e-14, disp=False,
            with_constraint=True):
        r"""
        Optimize coefficients and/or exponents of Gaussian basis functions with constraint.

        Parameters
        ----------
        c0 : ndarray
            Initial guess for coefficients of Gaussian basis functions.
        e0 : ndarray
            Initial guess for exponents of Gaussian basis functions.
        opt_coeffs : bool, optional
            Whether to optimize coefficients of Gaussian basis functions.
        opt_expons : bool, optional
            Whether to optimize exponents of Gaussian basis functions.
        maxiter : int, optional
            Maximum number of iterations.
        tol : float, optional
            For slsqp. precision goal for the value of objective function in the stopping criterion.
            For trust-constr, it is precision goal for the change in the variables.
        disp : bool
            If True, then it will print the iteration errors, convergence messages from the
            optimizer and will print out various error measures at each iteration.
        with_constraint : bool
            If true, then adds the constraint that the integration of the model density must
            be equal to the constraint of true density. The default is True.

        Returns
        -------
        dict :
            The optimization results presented as a dictionary containing:

            "coeffs" : ndarray
                The optimized coefficients of the Gaussian model.
            "exps" : ndarray
                The optimized exponents of the Gaussian model.
            "success": bool
                Whether the optimization exited successfully.
            "message" : str
                Information about the cause of termination.
            "fun" : float
                Values of KL divergence (objective function) at the final iteration.
            "jacobian": ndarray
                The Jacobian of the objective function w.r.t. coefficients and exponents.
            "performance" : list
                Values of various performance measures of modeled density at each iteration,
                as computed by `_BaseFit.goodness_of_fit` method.
            "time" : float
                The time in seconds it took to optimize.

        Notes
        -----
        - The coefficients and exponents are bounded to be positive.

        """
        # set bounds, initial guess & args
        if opt_coeffs and opt_expons:
            bounds = [(1.e-12, np.inf)] * 2 * self.model.nbasis
            x0 = np.concatenate((c0, e0))
            args = ()
        elif opt_coeffs:
            bounds = [(1.e-12, np.inf)] * self.model.nbasis
            x0 = c0
            args = ("fixed_expons", e0)
        elif opt_expons:
            bounds = [(1.e-12, np.inf)] * self.model.nbasis
            x0 = e0
            args = ("fixed_coeffs", c0)
        else:
            raise ValueError("Nothing to optimize!")
        # set constraints
        constraints = []
        if with_constraint:
            if self.method == "slsqp":
                constraints = [{"fun": self.const_norm, "type": "eq", "args": args}]
            elif self.method == "trust-constr":
                constraints = [NonlinearConstraint(self.const_norm, 0.0, 0.0, jac="3-point")]
        # set optimization options
        if self.method == "slsqp":
            options = {"ftol": tol, "maxiter": maxiter, "disp": disp}
        elif self.method == "trust-constr":
            # If the display is true then increase verbosity.
            verbose = 0
            if disp:
                verbose = 3
            options = {"xtol": tol, "maxiter": maxiter, "disp": disp, "verbose": verbose}
        # Set callback to computing the error measures we care about
        callback = None
        if disp:
            if self.method == "slsqp":
                # Template for the header.
                print("-" * (10 + 12 + 15 + 15 + 15 + 12))
                # Format is {identifier:width}, ^ means center it, | means put a bar in it
                template_header = (
                    "|{0:^12}|{1:^15}|{2:^15}|{3:^15}|{4:^16}|"
                )
                # Print the headers
                print(template_header.format(
                    "Integration", "L1", "L Infinity", "Least-squares", "Kullback-Leibler")
                )
                print("-" * (10 + 12 + 15 + 15 + 15 + 12))
                # Template for float iteration
                template_iters = (
                    "|{0:^12f}|{1:^15e}|{2:^15f}|{3:^15e}|{4:^16e}|"
                )

                def print_results(xk):
                    if opt_coeffs and opt_expons:
                        performance = self.goodness_of_fit(xk[:len(c0)], xk[len(c0):])
                    elif opt_coeffs:
                        performance = self.goodness_of_fit(xk, e0)
                    elif opt_expons:
                        performance = self.goodness_of_fit(c0, xk)
                    print(template_iters.format(*performance))

                callback = print_results
            elif self.method == "trust-constr":
                # Template for float iteration
                template_iters = (
                    "Integration: {0:10f},  L1: {1:10e},  L Infinity: {2:10f}, "
                    "Least Squares: {3:10e},  Kullback-Liebler: {4:10e}"
                )

                def print_results(xk, _):
                    if opt_coeffs and opt_expons:
                        performance = self.goodness_of_fit(xk[:len(c0)], xk[len(c0):])
                    elif opt_coeffs:
                        performance = self.goodness_of_fit(xk, e0)
                    elif opt_expons:
                        performance = self.goodness_of_fit(c0, xk)
                    print(template_iters.format(*performance))
                callback = print_results
        # optimize
        # scipy.minimize no longer supports extended precision
        x0 = x0.astype(np.float64)
        start = timer()  # Start timer
        res = minimize(fun=self.func,
                       x0=x0,
                       args=args,
                       method=self.method,
                       jac=True,
                       bounds=bounds,
                       constraints=constraints,
                       options=options,
                       callback=callback,
                       )
        end = timer()
        time = end - start

        # check successful optimization
        if not res["success"]:
            warnings.warn(f"Failed Optimization: {res['message']}")

        # split optimized coeffs & expons
        if opt_coeffs and opt_expons:
            coeffs, expons = res["x"][:self.model.nbasis], res["x"][self.model.nbasis:]
        elif opt_coeffs:
            coeffs, expons = res["x"], e0
        else:
            coeffs, expons = c0, res["x"]

        if disp:
            # Print Final Output.
            if self.method == "slsqp":
                print("-" * (10 + 12 + 15 + 15 + 15 + 12))
            print(f"Successful?: {res['success']}")
            print(f"Time: {time:.2f} seconds")
            print("Scipy Message: " + res["message"])
            print(f"Coefficients {coeffs}")
            print(f"Exponents {expons}")

        results = {"coeffs": coeffs,
                   "exps": expons,
                   "fun": res["fun"],
                   "success": res["success"],
                   "message": res["message"],
                   "jacobian": res["jac"],
                   "performance": np.array(self.goodness_of_fit(coeffs, expons)),
                   "time": time}
        return results

    def func(self, x, *args):
        r"""Compute objective function and its derivative w.r.t. Gaussian basis parameters.

        Parameters
        ----------
        x : ndarray
            The parameters of Gaussian basis which is being optimized. Contains both the
            coefficients and exponents together in a 1-D array.
        args :
            Additional arguments to the model.

        Returns
        -------
        (float, ndarray) :
            The objective function value and its derivative wrt to coefficients and exponents.

        """
        # compute linear combination of gaussian basis functions
        m, dm = self.evaluate_model(x, *args)
        # compute KL divergence
        k, dk = self.measure.evaluate(self.density, m, deriv=True)
        # compute objective function & its derivative
        obj = self.integrate(self.weights * k)
        d_obj = np.zeros_like(x)
        for index in range(len(x)):
            d_obj[index] = self.integrate(self.weights * dk * dm[:, index])
        return obj, d_obj

    def const_norm(self, x, *args):
        r"""Compute deviation in normalization constraint :math:`\sum c_i - \int f(x) dx`.

        Parameters
        ----------
        x : ndarray
            The parameters of Gaussian basis-functions. Contains both the
            coefficients and exponents together in a 1-D array.
        args :
            Additional parameters for the model.

        Returns
        -------
        float :
            The deviation of the integrla with the normalization constant.

        """
        # compute linear combination of gaussian basis functions
        m, _ = self.evaluate_model(x, *args)
        cons = self.integral_dens - self.integrate(m)
        return cons

    def evaluate_model(self, x, *args):
        r"""
        Evaluate the model density & its derivative.

        Parameters
        ----------
        x : ndarray
            The parameters of Gaussian basis-functions. Contains both the
            coefficients and exponents together in a 1-D array.
        args :
            Additional parameters for the model.

        Returns
        -------
        float, ndarray :
            Evaluates the model density & its derivative.

        """
        # assign coefficients & exponents
        if len(args) != 0:
            if args[0] == "fixed_coeffs":
                coeffs = args[1]
                expons = x
                start, end = self.model.nbasis, 2 * self.model.nbasis
            elif args[0] == "fixed_expons":
                coeffs = x
                expons = args[1]
                start, end = 0, self.model.nbasis
            else:
                raise ValueError("Argument args is not understandable!")
        else:
            coeffs, expons = x[:self.model.nbasis], x[self.model.nbasis:]
            start, end = 0, 2 * self.model.nbasis
        # compute model density & its derivative
        m, dm = self.model.evaluate(coeffs, expons, deriv=True)
        return m, dm[:, start: end]
